"""Full Text Operator."""

import os
from typing import List, Optional

from dbgpt.core import Chunk
from dbgpt.core.awel import MapOperator
from dbgpt.core.awel.flow import IOField, OperatorCategory, Parameter, ViewMetadata
from dbgpt.storage.full_text.base import FullTextStoreBase
from dbgpt.util.i18n_utils import _


class FullTextStorageOperator(MapOperator[List[Chunk], List[Chunk]]):
    """Full Text Operator."""

    metadata = ViewMetadata(
        label=_("Full Text Storage Operator"),
        name="full text_storage_operator",
        description=_("Persist embeddings into full text storage."),
        category=OperatorCategory.RAG,
        parameters=[
            Parameter.build_from(
                _("Full Text Connector"),
                "full_text_store",
                FullTextStoreBase,
                description=_("The full text store."),
                alias=["full_text_store"],
            ),
        ],
        inputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_("The text split chunks by chunk manager."),
                is_list=True,
            )
        ],
        outputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_(
                    "The assembled chunks, it has been persisted to full text store."
                ),
                is_list=True,
            )
        ],
    )

    def __init__(
        self,
        full_text_store: Optional[FullTextStoreBase] = None,
        max_chunks_once_load: Optional[int] = None,
        **kwargs,
    ):
        """Init the datasource operator."""
        MapOperator.__init__(self, **kwargs)
        self._full_text_store = full_text_store
        self._embeddings = full_text_store.get_config().embedding_fn
        self._max_chunks_once_load = max_chunks_once_load
        self.full_text_store = full_text_store

    async def map(self, chunks: List[Chunk]) -> List[Chunk]:
        """Persist chunks in full text db."""
        max_chunks_once_load = self._max_chunks_once_load or int(
            os.getenv("KNOWLEDGE_MAX_CHUNKS_ONCE_LOAD", 10)
        )
        full_text_ids = await self._full_text_store.aload_document_with_limit(
            chunks, max_chunks_once_load
        )
        for chunk, full_text_id in zip(chunks, full_text_ids):
            chunk.chunk_id = str(full_text_id)
        return chunks



"""S3 compatible storage backend."""

import hashlib
import io
import logging
import os
import random
import time
from typing import BinaryIO, Callable, Dict, Optional, Union
from urllib.parse import parse_qs, urlparse

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError

from dbgpt.core.interface.file import FileMetadata, StorageBackend

logger = logging.getLogger(__name__)


class S3Storage(StorageBackend):
    """S3 compatible storage backend implementation."""

    storage_type: str = "s3"

    def __init__(
        self,
        endpoint_url: str,
        region_name: str,
        access_key_id: str,
        secret_access_key: str,
        save_chunk_size: int = 1024 * 1024,
        use_environment_credentials: bool = False,
        fixed_bucket: Optional[str] = None,
        bucket_prefix: str = "dbgpt-fs-",
        bucket_mapper: Optional[Callable[[str], str]] = None,
        auto_create_bucket: bool = True,
        signature_version: Optional[str] = None,
        s3_config: Optional[Dict[str, Union[str, int]]] = None,
        default_public_url_expire: int = 3600,  # Default to 1 hour
    ):
        """Initialize the S3 compatible storage backend.

        Args:
            endpoint_url (str): S3 endpoint URL, e.g.,
                "https://s3.us-east-1.amazonaws.com"
            region_name (str): S3 region, e.g., "us-east-1"
            access_key_id (str): AWS/S3 Access Key ID
            secret_access_key (str): AWS/S3 Secret Access Key
            save_chunk_size (int, optional): Chunk size for saving files. Defaults to
                1024*1024 (1MB).
            use_environment_credentials (bool, optional): Whether to use credentials
                from environment variables. Defaults to False.
            fixed_bucket (Optional[str], optional): A fixed S3 bucket to use for all
                operations. If provided, all logical buckets will be mapped to this
                single bucket. Defaults to None.
            bucket_prefix (str, optional): Prefix for dynamically created buckets.
                Defaults to "dbgpt-fs-".
            bucket_mapper (Optional[Callable[[str], str]], optional): Custom function
                to map logical bucket names to actual S3 bucket names. Defaults to None.
            auto_create_bucket (bool, optional): Whether to automatically create
                buckets that don't exist. Defaults to True.
            signature_version (str, optional): S3 signature version to use.
            s3_config (Optional[Dict[str, Union[str, int]]], optional): Additional
                S3 configuration options. Defaults to None.
            default_public_url_expire (int, optional): Default expiration time for
                public URL in seconds. Defaults to 3600 (1 hour).
        """
        self.endpoint_url = endpoint_url
        self.region_name = region_name
        self._save_chunk_size = save_chunk_size
        self.fixed_bucket = fixed_bucket
        self.bucket_prefix = bucket_prefix
        self.custom_bucket_mapper = bucket_mapper
        self.auto_create_bucket = auto_create_bucket
        self.signature_version = signature_version
        self.default_public_url_expire = default_public_url_expire

        # Build S3 client configuration
        if not s3_config:
            s3_config = {
                "s3": {
                    # Use virtual addressing style
                    "addressing_style": "virtual",
                },
                "signature_version": signature_version or "v4",
            }
        if "request_checksum_calculation" not in s3_config:
            s3_config["request_checksum_calculation"] = "when_required"
        if "response_checksum_validation" not in s3_config:
            s3_config["response_checksum_validation"] = "when_required"
        config = Config(**s3_config)

        # Initialize S3 authentication
        if use_environment_credentials:
            # Check required environment variables
            required_env_vars = ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"]
            for var in required_env_vars:
                if var not in os.environ:
                    raise ValueError(f"Environment variable {var} is not set.")

            # Use environment credentials
            self.s3_client = boto3.client(
                "s3",
                endpoint_url=self.endpoint_url,
                region_name=self.region_name,
                config=config,
            )
        else:
            if not access_key_id or not secret_access_key:
                raise ValueError(
                    "Access key ID and secret are required when not using environment "
                    "credentials"
                )
            # Use provided credentials
            self.s3_client = boto3.client(
                "s3",
                endpoint_url=self.endpoint_url,
                region_name=self.region_name,
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key,
                config=config,
            )

        # Create fixed bucket if specified
        if self.fixed_bucket and self.auto_create_bucket:
            self._ensure_bucket_exists(self.fixed_bucket)

    @property
    def save_chunk_size(self) -> int:
        """Get the save chunk size."""
        return self._save_chunk_size

    def _map_bucket_name(self, logical_bucket: str) -> str:
        """Map logical bucket name to actual S3 bucket name.

        Args:
            logical_bucket (str): Logical bucket name used by the application

        Returns:
            str: Actual S3 bucket name to use
        """
        # 1. If using a fixed bucket, always return that
        if self.fixed_bucket:
            return self.fixed_bucket

        # 2. If a custom mapper is provided, use that
        if self.custom_bucket_mapper:
            return self.custom_bucket_mapper(logical_bucket)

        # 3. Otherwise, use a hash-based approach to generate a unique but
        # deterministic name
        # This avoids bucket name conflicts while maintaining consistency
        bucket_hash = hashlib.md5(logical_bucket.encode()).hexdigest()[:8]
        return f"{self.bucket_prefix}{bucket_hash}-{logical_bucket}"

    def _generate_dynamic_bucket_name(self) -> str:
        """Generate a unique bucket name for dynamic creation.

        Returns:
            str: A unique bucket name
        """
        # Using timestamp + random number to ensure uniqueness
        timestamp = int(time.time())
        random_number = random.randint(0, 9999)
        return f"{self.bucket_prefix}{timestamp}-{random_number}"

    def _ensure_bucket_exists(self, bucket_name: str) -> bool:
        """Ensure the bucket exists, create it if needed and if auto_create_bucket is
        True.

        Args:
            bucket_name (str): Bucket name

        Returns:
            bool: True if the bucket exists or was created, False otherwise
        """
        try:
            # Check if bucket exists
            self.s3_client.head_bucket(Bucket=bucket_name)
            logger.info(f"Bucket {bucket_name} exists")
            return True
        except ClientError as e:
            error_code = e.response.get("Error", {}).get("Code")
            error_msg = str(e)

            logger.info(
                f"Bucket check failed with error_code={error_code}, msg={error_msg}"
            )

            # Bucket doesn't exist or we don't have permission to access it
            if error_code in ["404", "403", "NoSuchBucket", "Forbidden"]:
                if not self.auto_create_bucket:
                    logger.warning(
                        f"Bucket {bucket_name} does not exist and auto_create_bucket "
                        "is False"
                    )
                    return False

                # Create bucket
                try:
                    logger.info(f"Creating bucket {bucket_name}")

                    # Try different creation methods to adapt to different
                    # S3-compatible APIs
                    creation_methods = [
                        # Method 1: Use LocationConstraint
                        lambda: self.s3_client.create_bucket(
                            Bucket=bucket_name,
                            CreateBucketConfiguration={
                                "LocationConstraint": self.region_name
                            },
                        ),
                        # Method 2: Without LocationConstraint
                        lambda: self.s3_client.create_bucket(Bucket=bucket_name),
                        # Method 3: Use empty CreateBucketConfiguration
                        lambda: self.s3_client.create_bucket(
                            Bucket=bucket_name, CreateBucketConfiguration={}
                        ),
                    ]

                    # Try different creation methods
                    last_error = None
                    for create_method in creation_methods:
                        try:
                            create_method()
                            logger.info(f"Successfully created bucket {bucket_name}")
                            return True
                        except ClientError as method_error:
                            logger.info(
                                f"Bucket creation method failed: {method_error}"
                            )
                            last_error = method_error
                            continue

                    # If all methods failed, raise the last error
                    if last_error:
                        raise last_error

                    return False

                except ClientError as create_error:
                    # Handle the case where bucket name is already taken by someone else
                    logger.error(
                        f"Failed to create bucket {bucket_name}: {create_error}"
                    )
                    if "BucketAlreadyExists" in str(create_error):
                        logger.warning(
                            f"Bucket name {bucket_name} already exists and is owned by "
                            "someone else"
                        )
                    return False
            else:
                # Some other error
                logger.error(f"Failed to check bucket {bucket_name}: {e}")
                return False

    def save(
        self,
        bucket: str,
        file_id: str,
        file_data: BinaryIO,
        public_url: bool = False,
        public_url_expire: Optional[int] = None,
    ) -> str:
        """Save the file data to S3.

        Args:
            bucket (str): The logical bucket name
            file_id (str): The file ID
            file_data (BinaryIO): The file data
            public_url (bool, optional): Whether to generate a public URL. Defaults to
                False.
            public_url_expire (Optional[int], optional): Expiration time for the public
                URL in seconds. Defaults to None.

        Returns:
            str: The storage path (S3 URI) or the public URL if public_url is True
        """
        # Get the actual S3 bucket
        actual_bucket_name = self._map_bucket_name(bucket)
        logger.info(
            f"Mapped logical bucket '{bucket}' to actual bucket '{actual_bucket_name}'"
        )

        # Ensure bucket exists
        bucket_exists = self._ensure_bucket_exists(actual_bucket_name)

        if not bucket_exists:
            logger.warning(
                f"Could not ensure bucket {actual_bucket_name} exists, trying "
                "alternatives"
            )

            # Try to create a dynamic bucket if we're not using a fixed bucket
            if not self.fixed_bucket and self.auto_create_bucket:
                dynamic_bucket = self._generate_dynamic_bucket_name()
                logger.info(
                    f"Attempting to create dynamic bucket {dynamic_bucket} for logical "
                    f"bucket {bucket}"
                )

                if self._ensure_bucket_exists(dynamic_bucket):
                    logger.info(f"Successfully created dynamic bucket {dynamic_bucket}")
                    actual_bucket_name = dynamic_bucket
                else:
                    error_msg = (
                        f"Failed to get or create bucket for logical bucket {bucket}"
                    )
                    logger.error(error_msg)
                    raise ValueError(error_msg)
            else:
                error_msg = (
                    f"Failed to get or create bucket for logical bucket {bucket}"
                )
                logger.error(error_msg)
                raise ValueError(error_msg)

        # Generate S3 object key based on whether we're using fixed bucket
        object_key = file_id
        if self.fixed_bucket:
            # When using a fixed bucket, we need to prefix with logical bucket name to
            # avoid conflicts
            object_key = f"{bucket}/{file_id}"

        # For large files, use multipart upload
        file_size = self._get_file_size(file_data)

        if file_size > 10 * self.save_chunk_size:  # If file is larger than 10MB
            logger.info(
                f"Using multipart upload for large file: {object_key} "
                f"(size: {file_size})"
            )
            self._multipart_upload(actual_bucket_name, object_key, file_data)
        else:
            logger.info(f"Uploading file using simple upload: {object_key}")
            try:
                # Reset the file pointer to the beginning
                file_data.seek(0)

                # Read the file content into memory
                file_content = file_data.read()

                # Use put_object for small files
                self.s3_client.put_object(
                    Bucket=actual_bucket_name, Key=object_key, Body=file_content
                )
            except ClientError as e:
                logger.error(
                    f"Failed to upload file {object_key} to bucket "
                    f"{actual_bucket_name}: {e}"
                )
                raise

        # Standard storage path
        storage_path = f"s3://{bucket}/{file_id}?actual_bucket={actual_bucket_name}&object_key={object_key}"  # noqa

        # Generate a public URL if requested
        if public_url:
            # Use provided expiration time or default
            expire_seconds = public_url_expire or self.default_public_url_expire

            try:
                # Generate a pre-signed URL for public access
                url = self.s3_client.generate_presigned_url(
                    "get_object",
                    Params={"Bucket": actual_bucket_name, "Key": object_key},
                    ExpiresIn=expire_seconds,
                )
                logger.info(
                    f"Generated public URL for {object_key} with expiration "
                    f"{expire_seconds} seconds"
                )
                return url
            except ClientError as e:
                logger.error(f"Failed to generate public URL for {object_key}: {e}")
                # Fall back to returning the storage path
                return storage_path

        return storage_path

    def _get_file_size(self, file_data: BinaryIO) -> int:
        """Get file size without consuming the file object.

        Args:
            file_data (BinaryIO): The file data

        Returns:
            int: The file size in bytes
        """
        current_pos = file_data.tell()
        file_data.seek(0, io.SEEK_END)
        size = file_data.tell()
        file_data.seek(current_pos)  # Reset the file pointer
        return size

    def _multipart_upload(
        self, bucket_name: str, object_key: str, file_data: BinaryIO
    ) -> None:
        """Handle multipart upload for large files.

        Args:
            bucket_name (str): S3 bucket name
            object_key (str): The object key (file path in S3)
            file_data (BinaryIO): The file data
        """
        # Initialize multipart upload
        try:
            mpu = self.s3_client.create_multipart_upload(
                Bucket=bucket_name, Key=object_key
            )
            upload_id = mpu["UploadId"]

            # Upload parts
            part_number = 1
            parts = []
            file_data.seek(0)  # Make sure we're at the beginning of the file

            while True:
                # Read the chunk
                chunk = file_data.read(self.save_chunk_size)
                if not chunk:
                    break

                # Upload the part
                response = self.s3_client.upload_part(
                    Bucket=bucket_name,
                    Key=object_key,
                    UploadId=upload_id,
                    PartNumber=part_number,
                    Body=chunk,
                )

                parts.append({"PartNumber": part_number, "ETag": response["ETag"]})

                part_number += 1

            # Complete multipart upload
            self.s3_client.complete_multipart_upload(
                Bucket=bucket_name,
                Key=object_key,
                UploadId=upload_id,
                MultipartUpload={"Parts": parts},
            )
        except ClientError as e:
            logger.error(f"Error in multipart upload: {e}")
            # Attempt to abort the multipart upload if it was initialized
            if "upload_id" in locals():
                try:
                    self.s3_client.abort_multipart_upload(
                        Bucket=bucket_name, Key=object_key, UploadId=upload_id
                    )
                except ClientError as abort_error:
                    logger.error(f"Error aborting multipart upload: {abort_error}")
            raise

    def _parse_storage_path(self, storage_path: str) -> Dict[str, str]:
        """Parse the S3 storage path to extract actual bucket and object key.

        Args:
            storage_path (str): The storage path URI

        Returns:
            Dict[str, str]: A dictionary with actual_bucket and object_key keys
        """
        if not storage_path.startswith("s3://"):
            raise ValueError(f"Invalid storage path for S3: {storage_path}")

        # Example URI:
        # s3://logical_bucket/file_id?actual_bucket=s3_bucket&object_key=logical_bucket/file_id # noqa

        # Parse the URL
        parsed_url = urlparse(storage_path)
        params = parse_qs(parsed_url.query)

        # Extract the parameters
        actual_bucket = params.get("actual_bucket", [None])[0]
        object_key = params.get("object_key", [None])[0]

        # Extract the logical bucket and file_id from the path
        path_parts = parsed_url.path.strip("/").split("/", 1)
        logical_bucket = path_parts[0] if path_parts else None
        logical_file_id = path_parts[1] if len(path_parts) > 1 else None

        # If parameters aren't in the URL (backward compatibility or simplified URL),
        # derive them from the logical values
        if not actual_bucket:
            # Try to use the bucket mapper to get the actual bucket
            actual_bucket = (
                self._map_bucket_name(logical_bucket) if logical_bucket else None
            )

        if not object_key:
            # If using fixed bucket, the object key includes the logical bucket
            # as prefix
            if self.fixed_bucket:
                object_key = (
                    f"{logical_bucket}/{logical_file_id}"
                    if logical_bucket and logical_file_id
                    else None
                )
            else:
                object_key = logical_file_id

        return {
            "logical_bucket": logical_bucket,
            "logical_file_id": logical_file_id,
            "actual_bucket": actual_bucket,
            "object_key": object_key,
        }

    def get_public_url(
        self, fm: FileMetadata, expire: Optional[int] = None
    ) -> Optional[str]:
        """Generate a public URL for an existing file.

        Args:
            fm (FileMetadata): The file metadata
            expire (Optional[int], optional): Expiration time in seconds. Defaults to
                class default.

        Returns:
            str: The generated public URL
        """
        # Parse the storage path
        path_info = self._parse_storage_path(fm.storage_path)

        # Get actual bucket and object key
        actual_bucket_name = path_info["actual_bucket"]
        object_key = path_info["object_key"]
        logical_bucket = path_info["logical_bucket"]

        # If we couldn't determine the actual bucket from the URI, try with the logical
        #  bucket
        if not actual_bucket_name and logical_bucket:
            actual_bucket_name = self._map_bucket_name(logical_bucket)

        # Use the file_id as object key if object_key is still None
        if not object_key:
            object_key = fm.file_id
            # If using fixed bucket, prefix with logical bucket
            if self.fixed_bucket and logical_bucket:
                object_key = f"{logical_bucket}/{fm.file_id}"

        # Use provided expiration time or default
        expire_seconds = expire or self.default_public_url_expire

        try:
            # Generate a pre-signed URL for public access
            url = self.s3_client.generate_presigned_url(
                "get_object",
                Params={"Bucket": actual_bucket_name, "Key": object_key},
                ExpiresIn=expire_seconds,
            )
            logger.info(
                f"Generated public URL for {object_key} with expiration "
                f"{expire_seconds} seconds"
            )
            return url
        except ClientError as e:
            logger.error(f"Failed to generate public URL for {fm.file_id}: {e}")
            raise

    def load(self, fm: FileMetadata) -> BinaryIO:
        """Load the file data from S3.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            BinaryIO: The file data as a binary IO object
        """
        # Parse the storage path
        path_info = self._parse_storage_path(fm.storage_path)

        # Get actual bucket and object key
        actual_bucket_name = path_info["actual_bucket"]
        object_key = path_info["object_key"]
        logical_bucket = path_info["logical_bucket"]

        # If we couldn't determine the actual bucket from the URI, try with the
        # logical bucket
        if not actual_bucket_name and logical_bucket:
            actual_bucket_name = self._map_bucket_name(logical_bucket)

        # Use the file_id as object key if object_key is still None
        if not object_key:
            object_key = fm.file_id
            # If using fixed bucket, prefix with logical bucket
            if self.fixed_bucket and logical_bucket:
                object_key = f"{logical_bucket}/{fm.file_id}"

        try:
            # Get object from S3
            response = self.s3_client.get_object(
                Bucket=actual_bucket_name, Key=object_key
            )

            # Read the streaming body into a BytesIO object
            content = io.BytesIO()
            body = response["Body"]

            # Stream the data in chunks
            while True:
                chunk = body.read(self.save_chunk_size)
                if not chunk:
                    break
                content.write(chunk)

            content.seek(0)
            return content
        except ClientError as e:
            error_code = e.response.get("Error", {}).get("Code")
            if error_code == "NoSuchKey":
                logger.error(
                    f"File {object_key} not found in bucket {actual_bucket_name}: {e}"
                )
                raise FileNotFoundError(
                    f"File {object_key} not found in bucket {actual_bucket_name}"
                )
            logger.error(
                f"Failed to download file {object_key} from bucket "
                f"{actual_bucket_name}: {e}"
            )
            raise

    def delete(self, fm: FileMetadata) -> bool:
        """Delete the file data from S3.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            bool: True if the file was deleted, False otherwise
        """
        # Parse the storage path
        path_info = self._parse_storage_path(fm.storage_path)

        # Get actual bucket and object key
        actual_bucket_name = path_info["actual_bucket"]
        object_key = path_info["object_key"]
        logical_bucket = path_info["logical_bucket"]

        # If we couldn't determine the actual bucket from the URI, try with the
        # logical bucket
        if not actual_bucket_name and logical_bucket:
            actual_bucket_name = self._map_bucket_name(logical_bucket)

        # Use the file_id as object key if object_key is still None
        if not object_key:
            object_key = fm.file_id
            # If using fixed bucket, prefix with logical bucket
            if self.fixed_bucket and logical_bucket:
                object_key = f"{logical_bucket}/{fm.file_id}"

        try:
            # Check if the object exists
            try:
                self.s3_client.head_object(Bucket=actual_bucket_name, Key=object_key)
            except ClientError as e:
                error_code = e.response.get("Error", {}).get("Code")
                if error_code == "404" or error_code == "NoSuchKey":
                    logger.warning(
                        f"File {object_key} does not exist in bucket "
                        f"{actual_bucket_name}"
                    )
                    return False
                raise

            # Delete the object
            self.s3_client.delete_object(Bucket=actual_bucket_name, Key=object_key)

            logger.info(f"File {object_key} deleted from bucket {actual_bucket_name}")
            return True
        except ClientError as e:
            logger.error(
                f"Failed to delete file {object_key} from bucket {actual_bucket_name}:"
                f" {e}"
            )
            return False



"""MSSQL connector."""

from dataclasses import dataclass, field
from typing import Iterable, Type

from sqlalchemy import text

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.rdbms.base import RDBMSConnector, RDBMSDatasourceParameters
from dbgpt.util.i18n_utils import _


@auto_register_resource(
    label=_("MSSQL datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_(
        "Powerful, scalable, secure relational database system by Microsoft."
    ),
)
@dataclass
class MSSQLParameters(RDBMSDatasourceParameters):
    """MSSQL connection parameters."""

    __type__ = "mssql"
    driver: str = field(
        default="mssql+pymssql",
        metadata={
            "help": _("Driver name for MSSQL, default is mssql+pymssql."),
        },
    )

    def create_connector(self) -> "MSSQLConnector":
        """Create MS SQL connector."""
        return MSSQLConnector.from_parameters(self)


class MSSQLConnector(RDBMSConnector):
    """MSSQL connector."""

    db_type: str = "mssql"
    db_dialect: str = "mssql"
    driver: str = "mssql+pymssql"

    default_db = ["master", "model", "msdb", "tempdb", "modeldb", "resource", "sys"]

    @classmethod
    def param_class(cls) -> Type[MSSQLParameters]:
        """Return the parameter class."""
        return MSSQLParameters

    def table_simple_info(self) -> Iterable[str]:
        """Get table simple info."""
        _tables_sql = """
                SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE
                TABLE_TYPE='BASE TABLE'
            """
        with self.session_scope() as session:
            cursor = session.execute(text(_tables_sql))
            tables_results = cursor.fetchall()
            results = []
            for row in tables_results:
                table_name = row[0]
                _sql = f"""
                    SELECT COLUMN_NAME, DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS WHERE
                     TABLE_NAME='{table_name}'
                """
                cursor_colums = session.execute(text(_sql))
                colum_results = cursor_colums.fetchall()
                table_colums = []
                for row_col in colum_results:
                    field_info = list(row_col)
                    table_colums.append(field_info[0])
                results.append(f"{table_name}({','.join(table_colums)});")
            return results

    def get_users(self):
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    "SELECT name FROM sys.server_principals "
                    "WHERE type_desc = 'SQL_LOGIN'"
                )
            )
            return [row[0] for row in cursor.fetchall()]

    def get_grants(self):
        with self.session_scope() as session:
            query = """
            SELECT 
                CASE WHEN perm.state <> 'W' THEN perm.state_desc ELSE 'GRANT WITH
                 GRANT OPTION' END AS [Permission],
                perm.permission_name AS [Permission Name],
                CASE 
                    WHEN perm.class = 0 THEN 'SERVER'
                    WHEN perm.class = 1 THEN OBJECT_NAME(perm.major_id)
                    WHEN perm.class = 3 THEN SCHEMA_NAME(perm.major_id) 
                    ELSE CAST(perm.class AS VARCHAR)
                END AS [Securable],
                princ.name AS [Principal]
            FROM 
                sys.server_permissions perm
                JOIN sys.server_principals princ ON perm.grantee_principal_id = 
                princ.principal_id
            """
            cursor = session.execute(text(query))
            return cursor.fetchall()

    def _decode_if_bytes(self, value):
        if isinstance(value, bytes):
            return value.decode("utf-8")
        return value

    def get_charset(self):
        with self.session_scope() as session:
            query = (
                "SELECT DATABASEPROPERTYEX(DB_NAME(), 'Collation') AS DatabaseCollation"
            )
            cursor = session.execute(text(query))
            result = cursor.fetchone()

            if result and result[0]:
                collation = self._decode_if_bytes(result[0])
                parts = collation.split("_")
                if len(parts) >= 2:
                    return parts[1]
                return collation

            return "SQL_Server_Default"

    def get_collation(self):
        with self.session_scope() as session:
            cursor = session.execute(
                text("SELECT SERVERPROPERTY('Collation') AS DatabaseCollation")
            )
            collation = cursor.fetchone()[0]
            return collation

    def get_table_names(self):
        tables = []

        with self.session_scope() as session:
            query = """
            SELECT 
                TABLE_SCHEMA + '.' + TABLE_NAME AS full_table_name
            FROM 
                INFORMATION_SCHEMA.TABLES
            WHERE 
                TABLE_TYPE = 'BASE TABLE'
                AND TABLE_CATALOG = DB_NAME()
            """

            cursor = session.execute(text(query))
            tables = [row[0] for row in cursor.fetchall()]

            if not tables:
                query = """
                SELECT 
                    SCHEMA_NAME(schema_id) + '.' + name AS full_table_name
                FROM 
                    sys.tables
                """
                cursor = session.execute(text(query))
                tables = [row[0] for row in cursor.fetchall()]

            if not tables:
                query = """
                SELECT 
                    name AS table_name
                FROM 
                    sys.tables
                """
                cursor = session.execute(text(query))
                tables = [row[0] for row in cursor.fetchall()]

            return tables

    def get_columns(self, table_name: str):
        if "." in table_name:
            schema_name, pure_table_name = table_name.split(".", 1)
        else:
            schema_name = "dbo"
            pure_table_name = table_name

        with self.session_scope() as session:
            query = """
            SELECT 
                COLUMN_NAME AS name,
                DATA_TYPE AS type,
                CASE WHEN IS_NULLABLE = 'YES' THEN 1 ELSE 0 END AS nullable,
                COLUMN_DEFAULT AS default_value,
                CHARACTER_MAXIMUM_LENGTH AS max_length
            FROM 
                INFORMATION_SCHEMA.COLUMNS
            WHERE 
                TABLE_SCHEMA = :schema
                AND TABLE_NAME = :table
            ORDER BY 
                ORDINAL_POSITION
            """
            cursor = session.execute(
                text(query), {"schema": schema_name, "table": pure_table_name}
            )
            results = cursor.fetchall()

            columns = []
            for row in results:
                name = row[0].decode("utf-8") if isinstance(row[0], bytes) else row[0]
                col_type = (
                    row[1].decode("utf-8") if isinstance(row[1], bytes) else row[1]
                )
                column = {
                    "name": name,
                    "type": col_type,
                    "nullable": bool(row[2]),
                }

                if row[3] is not None:
                    default = (
                        row[3].decode("utf-8") if isinstance(row[3], bytes) else row[3]
                    )
                    column["default"] = default

                if row[4] is not None:
                    column["max_length"] = row[4]
                columns.append(column)

            return columns

    def get_indexes(self, table_name: str):
        if "." in table_name:
            schema_name, pure_table_name = table_name.split(".", 1)
        else:
            schema_name = "dbo"
            pure_table_name = table_name

        with self.session_scope() as session:
            query = """
            SELECT
                i.name AS index_name,
                c.name AS column_name,
                i.is_unique AS is_unique,
                i.is_primary_key AS is_primary_key
            FROM
                sys.indexes i
            INNER JOIN
                sys.index_columns ic ON i.object_id = ic.object_id AND i.index_id 
                = ic.index_id
            INNER JOIN
                sys.columns c ON ic.object_id = c.object_id AND ic.column_id 
                = c.column_id
            INNER JOIN
                sys.tables t ON i.object_id = t.object_id
            INNER JOIN
                sys.schemas s ON t.schema_id = s.schema_id
            WHERE
                t.name = :table
                AND s.name = :schema
                AND i.name IS NOT NULL
            ORDER BY
                i.name, ic.key_ordinal
            """
            cursor = session.execute(
                text(query), {"schema": schema_name, "table": pure_table_name}
            )
            results = cursor.fetchall()

            index_dict = {}
            for row in results:
                index_name = (
                    row[0].decode("utf-8") if isinstance(row[0], bytes) else row[0]
                )
                column_name = (
                    row[1].decode("utf-8") if isinstance(row[1], bytes) else row[1]
                )
                is_unique = bool(row[2])
                is_primary_key = bool(row[3])
                if index_name not in index_dict:
                    index_dict[index_name] = {
                        "name": index_name,
                        "column_names": [],
                        "unique": is_unique,
                        "primary": is_primary_key,
                    }

                index_dict[index_name]["column_names"].append(column_name)

            return list(index_dict.values())



"""Module for embedding related classes and functions."""

from .aimlapi import AimlapiEmbeddings  # noqa: F401
from .jina import JinaEmbeddings  # noqa: F401
from .ollama import OllamaEmbeddings  # noqa: F401
from .qianfan import QianFanEmbeddings  # noqa: F401
from .siliconflow import SiliconFlowEmbeddings  # noqa: F401
from .tongyi import TongYiEmbeddings  # noqa: F401

__ALL__ = [
    "JinaEmbeddings",
    "OllamaEmbeddings",
    "QianFanEmbeddings",
    "TongYiEmbeddings",
    "SiliconFlowEmbeddings",
    "AimlapiEmbeddings",
]



"""Knowledge Graph Operator."""

import os
from typing import List, Optional

from dbgpt.core import Chunk
from dbgpt.core.awel import MapOperator
from dbgpt.core.awel.flow import IOField, OperatorCategory, Parameter, ViewMetadata
from dbgpt.storage.knowledge_graph.base import KnowledgeGraphBase
from dbgpt.util.i18n_utils import _


class KnowledgeGraphOperator(MapOperator[List[Chunk], List[Chunk]]):
    """Knowledge Graph Operator."""

    metadata = ViewMetadata(
        label=_("Knowledge Graph Operator"),
        name="knowledge_graph_operator",
        description=_("Extract Documents and persist into graph database."),
        category=OperatorCategory.RAG,
        parameters=[
            Parameter.build_from(
                _("Knowledge Graph Connector"),
                "graph_store",
                KnowledgeGraphBase,
                description=_("The knowledge graph."),
                alias=["graph_store"],
            ),
        ],
        inputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_("The text split chunks by chunk manager."),
                is_list=True,
            )
        ],
        outputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_(
                    "The assembled chunks, it has been persisted to graph store."
                ),
                is_list=True,
            )
        ],
    )

    def __init__(
        self,
        graph_store: Optional[KnowledgeGraphBase] = None,
        max_chunks_once_load: Optional[int] = None,
        **kwargs,
    ):
        """Init the Knowledge Graph operator."""
        MapOperator.__init__(self, **kwargs)
        self._graph_store = graph_store
        self._embeddings = graph_store.embeddings
        self._max_chunks_once_load = max_chunks_once_load
        self.graph_store = graph_store

    async def map(self, chunks: List[Chunk]) -> List[Chunk]:
        """Persist chunks in graph db."""
        max_chunks_once_load = self._max_chunks_once_load or int(
            os.getenv("KNOWLEDGE_MAX_CHUNKS_ONCE_LOAD", 10)
        )
        graph_ids = await self._graph_store.aload_document_with_limit(
            chunks, max_chunks_once_load
        )
        for chunk, graph_id in zip(chunks, graph_ids):
            chunk.chunk_id = str(graph_id)
        return chunks



"""Postgres vector store."""

import logging
import os
from dataclasses import dataclass, field
from typing import Any, List, Optional

from dbgpt.core import Chunk, Embeddings
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.vector_store.base import (
    _COMMON_PARAMETERS,
    _VECTOR_STORE_COMMON_PARAMETERS,
    VectorStoreBase,
    VectorStoreConfig,
)
from dbgpt.storage.vector_store.filters import MetadataFilters
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)


@register_resource(
    _("PGVector Config"),
    "pg_vector_config",
    category=ResourceCategory.VECTOR_STORE,
    parameters=[
        *_COMMON_PARAMETERS,
        Parameter.build_from(
            _("Connection String"),
            "connection_string",
            str,
            description=_(
                "The connection string of vector store, if not set, will use "
                "the default connection string."
            ),
            optional=True,
            default=None,
        ),
    ],
    description="PG vector config.",
)
@dataclass
class PGVectorConfig(VectorStoreConfig):
    """PG vector store config."""

    __type__ = "pgvector"

    connection_string: str = field(
        default=None,
        metadata={
            "description": "the connection string of vector store, "
            "if not set, will use the default connection string."
        },
    )

    def create_store(self, **kwargs) -> "PGVectorStore":
        """Create a PGVectorStore instance."""
        return PGVectorStore(vector_store_config=self, **kwargs)


@register_resource(
    _("PG Vector Store"),
    "pg_vector_store",
    category=ResourceCategory.VECTOR_STORE,
    description=_("PG vector store."),
    parameters=[
        Parameter.build_from(
            _("PG Config"),
            "vector_store_config",
            PGVectorConfig,
            description=_("the pg config of vector store."),
            optional=True,
            default=None,
        ),
        *_VECTOR_STORE_COMMON_PARAMETERS,
    ],
)
class PGVectorStore(VectorStoreBase):
    """PG vector store.

    To use this, you should have the ``pgvector`` python package installed.
    """

    def __init__(
        self,
        vector_store_config: PGVectorConfig,
        name: Optional[str],
        embedding_fn: Optional[Embeddings] = None,
        max_chunks_once_load: Optional[int] = None,
        max_threads: Optional[int] = None,
    ) -> None:
        """Create a PGVectorStore instance."""
        try:
            from langchain.vectorstores import PGVector  # mypy: ignore
        except ImportError:
            raise ImportError(
                "Please install the `langchain` package to use the PGVector."
            )
        super().__init__(
            max_chunks_once_load=max_chunks_once_load, max_threads=max_threads
        )
        self._vector_store_config = vector_store_config
        self.connection_string = vector_store_config.connection_string or os.getenv(
            "PGVECTOR_CONNECTION_STRING"
        )

        self.embeddings = embedding_fn
        self.collection_name = name

        self.vector_store_client = PGVector(
            embedding_function=self.embeddings,  # type: ignore
            collection_name=self.collection_name,
            connection_string=self.connection_string,
        )

    def get_config(self) -> PGVectorConfig:
        """Get the vector store config."""
        return self._vector_store_config

    def create_collection(self, collection_name: str, **kwargs) -> Any:
        """Create the collection."""
        return self.vector_store_client.create_collection()

    def similar_search(
        self, text: str, topk: int, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Perform similar search in PGVector."""
        return self.vector_store_client.similarity_search(text, topk, filters)

    def vector_name_exists(self) -> bool:
        """Check if vector name exists."""
        try:
            self.vector_store_client.create_collection()
            return True
        except Exception as e:
            logger.error(f"vector_name_exists error, {str(e)}")
            return False

    def load_document(self, chunks: List[Chunk]) -> List[str]:
        """Load document to PGVector.

        Args:
            chunks(List[Chunk]): document chunks.

        Return:
            List[str]: chunk ids.
        """
        lc_documents = [Chunk.chunk2langchain(chunk) for chunk in chunks]
        self.vector_store_client.from_documents(lc_documents)  # type: ignore
        return [str(chunk.chunk_id) for chunk in lc_documents]

    def delete_vector_name(self, vector_name: str):
        """Delete vector by name.

        Args:
            vector_name(str): vector name.
        """
        return self.vector_store_client.delete_collection()

    def delete_by_ids(self, ids: str):
        """Delete vector by ids.

        Args:
            ids(str): vector ids, separated by comma.
        """
        delete_ids = ids.split(",")
        return self.vector_store_client.delete(delete_ids)



"""Knowledge Process Branch Operator."""

from typing import Dict, List, Optional

from dbgpt.core import Chunk
from dbgpt.core.awel import (
    BranchFunc,
    BranchOperator,
    BranchTaskType,
    JoinOperator,
    logger,
)
from dbgpt.core.awel.flow import IOField, OperatorCategory, OperatorType, ViewMetadata
from dbgpt.rag.knowledge.base import Knowledge
from dbgpt.util.i18n_utils import _


class KnowledgeProcessBranchOperator(BranchOperator[Knowledge, Knowledge]):
    """Knowledge Process branch operator.

    This operator will branch the workflow based on
    the stream flag of the request.
    """

    metadata = ViewMetadata(
        label=_("Knowledge Process Branch Operator"),
        name="knowledge_process_operator",
        category=OperatorCategory.RAG,
        operator_type=OperatorType.BRANCH,
        description=_("Branch the workflow based on the stream flag of the request."),
        parameters=[],
        inputs=[
            IOField.build_from(
                _("Document Chunks"),
                "input_value",
                List[Chunk],
                description=_("The input value of the operator."),
                is_list=True,
            ),
        ],
        outputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_("Chunks for Vector Storage Connector."),
                is_list=True,
            ),
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_("Chunks for Knowledge Graph Connector."),
                is_list=True,
            ),
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_("Chunks for Full Text Connector."),
                is_list=True,
            ),
        ],
    )

    def __init__(
        self,
        graph_task_name: Optional[str] = None,
        embedding_task_name: Optional[str] = None,
        **kwargs,
    ):
        """Create the intent detection branch operator."""
        super().__init__(**kwargs)
        self._graph_task_name = graph_task_name
        self._embedding_task_name = embedding_task_name
        self._full_text_task_name = embedding_task_name

    async def branches(
        self,
    ) -> Dict[BranchFunc[Knowledge], BranchTaskType]:
        """Branch the intent detection result to different tasks."""
        download_cls_list = set(task.__class__ for task in self.downstream)  # noqa
        branch_func_map = {}

        async def check_graph_process(r: Knowledge) -> bool:
            # If check graph is true, we will run extract knowledge graph triplets.
            from dbgpt_ext.rag.operators import KnowledgeGraphOperator

            if KnowledgeGraphOperator in download_cls_list:
                return True
            return False

        async def check_embedding_process(r: Knowledge) -> bool:
            # If check embedding is true, we will run extract document embedding.
            from dbgpt_ext.rag.operators import VectorStorageOperator

            if VectorStorageOperator in download_cls_list:
                return True
            return False

        async def check_full_text_process(r: Knowledge) -> bool:
            # If check full text is true, we will run extract document keywords.
            from dbgpt_ext.rag.operators.full_text import FullTextStorageOperator

            if FullTextStorageOperator in download_cls_list:
                return True
            return False

        branch_func_map[check_graph_process] = self._graph_task_name
        branch_func_map[check_embedding_process] = self._embedding_task_name
        branch_func_map[check_full_text_process] = self._full_text_task_name
        return branch_func_map  # type: ignore


class KnowledgeProcessJoinOperator(JoinOperator[List[str]]):
    """Knowledge Process Join Operator."""

    metadata = ViewMetadata(
        label=_("Knowledge Process Join Operator"),
        name="knowledge_process_join_operator",
        category=OperatorCategory.RAG,
        operator_type=OperatorType.JOIN,
        description=_(
            "Join Branch the workflow based on the Knowledge Process Results."
        ),
        parameters=[],
        inputs=[
            IOField.build_from(
                _("Vector Storage Results"),
                "input_value",
                List[Chunk],
                description=_("vector storage results."),
                is_list=True,
            ),
            IOField.build_from(
                _("Knowledge Graph Storage Results"),
                "input_value",
                List[Chunk],
                description=_("knowledge graph storage results."),
                is_list=True,
            ),
        ],
        outputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_("Knowledge Process Results."),
                is_list=True,
            ),
        ],
    )

    def __init__(
        self,
        **kwargs,
    ):
        """Knowledge Process Join Operator."""
        super().__init__(combine_function=self._join, **kwargs)

    async def _join(
        self,
        vector_chunks: Optional[List[Chunk]] = None,
        graph_chunks: Optional[List[Chunk]] = None,
        full_text_chunks: Optional[List[Chunk]] = None,
    ) -> List[str]:
        """Join results.

        Args:
            vector_chunks: The list of vector chunks.
            graph_chunks: The list of graph chunks.
            full_text_chunks: The list of full text chunks.
        """
        results = []
        if vector_chunks:
            result_msg = (
                f"async persist vector store success {len(vector_chunks)} chunks."
            )
            logger.info(result_msg)
            results.append(result_msg)
        if graph_chunks:
            result_msg = (
                f"async persist graph store success {len(graph_chunks)} chunks."
            )
            logger.info(result_msg)
            results.append(result_msg)
        if full_text_chunks:
            result_msg = (
                f"async persist full text store success {len(full_text_chunks)} chunks."
            )
            logger.info(result_msg)
            results.append(result_msg)
        return results



"""Ollama Embeddings."""

from dataclasses import dataclass, field
from typing import List, Optional, Type

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field
from dbgpt.core import EmbeddingModelMetadata, Embeddings
from dbgpt.core.interface.parameter import EmbeddingDeployModelParameters
from dbgpt.model.adapter.base import register_embedding_adapter
from dbgpt.util.i18n_utils import _


@dataclass
class OllamaEmbeddingDeployModelParameters(EmbeddingDeployModelParameters):
    """Ollama Embeddings deploy model parameters."""

    provider: str = "proxy/ollama"

    api_url: str = field(
        default="http://localhost:11434",
        metadata={
            "help": _("The URL of the embeddings API."),
        },
    )
    backend: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The real model name to pass to the provider, default is None. If "
                "backend is None, use name as the real model name."
            ),
        },
    )

    @property
    def real_provider_model_name(self) -> str:
        """Get the real provider model name."""
        return self.backend or self.name


class OllamaEmbeddings(BaseModel, Embeddings):
    """Ollama proxy embeddings.

    This class is used to get embeddings for a list of texts using the Ollama API.
    It requires a proxy server url `api_url` and a model name `model_name`.
    The default model name is "llama2".
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())

    api_url: str = Field(
        default="http://localhost:11434",
        description="The URL of the embeddings API.",
    )
    model_name: str = Field(
        default="llama2", description="The name of the model to use."
    )

    def __init__(self, **kwargs):
        """Initialize the OllamaEmbeddings."""
        super().__init__(**kwargs)

    @classmethod
    def param_class(cls) -> Type[OllamaEmbeddingDeployModelParameters]:
        """Get the parameter class."""
        return OllamaEmbeddingDeployModelParameters

    @classmethod
    def from_parameters(
        cls, parameters: OllamaEmbeddingDeployModelParameters
    ) -> "Embeddings":
        """Create an embedding model from parameters."""
        return cls(
            api_url=parameters.api_url, model_name=parameters.real_provider_model_name
        )

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Get the embeddings for a list of texts.

        Args:
            texts (Documents): A list of texts to get embeddings for.

        Returns:
            Embedded texts as List[List[float]], where each inner List[float]
                corresponds to a single input text.
        """
        return [self.embed_query(text) for text in texts]

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a OpenAPI embedding model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        try:
            import ollama
            from ollama import Client
        except ImportError as e:
            raise ValueError(
                "Could not import python package: ollama "
                "Please install ollama by command `pip install ollama"
            ) from e
        try:
            embedding = Client(self.api_url).embeddings(
                model=self.model_name, prompt=text
            )
            return list(embedding["embedding"])
        except ollama.ResponseError as e:
            raise ValueError(f"**Ollama Response Error, Please CheckErrorInfo.**: {e}")

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Asynchronous Embed search docs.

        Args:
            texts: A list of texts to get embeddings for.

        Returns:
            List[List[float]]: Embedded texts as List[List[float]], where each inner
                List[float] corresponds to a single input text.
        """
        embeddings = []
        for text in texts:
            embedding = await self.aembed_query(text)
            embeddings.append(embedding)
        return embeddings

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous Embed query text."""
        try:
            import ollama
            from ollama import AsyncClient
        except ImportError:
            raise ValueError(
                "The ollama python package is not installed. "
                "Please install it with `pip install ollama`"
            )
        try:
            embedding = await AsyncClient(host=self.api_url).embeddings(
                model=self.model_name, prompt=text
            )
            return list(embedding["embedding"])
        except ollama.ResponseError as e:
            raise ValueError(f"**Ollama Response Error, Please CheckErrorInfo.**: {e}")


register_embedding_adapter(
    OllamaEmbeddings,
    supported_models=[
        EmbeddingModelMetadata(
            model=["BAAI/bge-m3"],
            dimension=1024,
            context_length=8192,
            description=_(
                "The embedding model are trained by BAAI, it support more than 100 "
                "working languages."
            ),
            link="https://ollama.com/library/bge-m3",
        )
    ],
)



"""TuGraph Community Store Adapter."""

import json
import logging
from typing import AsyncGenerator, Dict, Iterator, List, Literal, Optional, Tuple, Union

from dbgpt.storage.graph_store.graph import (
    Direction,
    Edge,
    Graph,
    GraphElemType,
    MemoryGraph,
    Vertex,
)
from dbgpt.storage.graph_store.memgraph_store import (
    MemoryGraphStore,
    MemoryGraphStoreConfig,
)
from dbgpt.storage.knowledge_graph.base import ParagraphChunk
from dbgpt_ext.storage.knowledge_graph.community.base import (
    Community,
    GraphStoreAdapter,
)

logger = logging.getLogger(__name__)


class MemGraphStoreAdapter(GraphStoreAdapter):
    """MemGraph Community Store Adapter."""

    MAX_HIERARCHY_LEVEL = 3

    def __init__(self, enable_summary: bool = False):
        """Initialize MemGraph Community Store Adapter."""
        self._graph_store: MemoryGraphStore = MemoryGraphStore(MemoryGraphStoreConfig())

        super().__init__(self._graph_store)

        # Create the graph
        self.create_graph(self._graph_store.get_config().name)

    async def discover_communities(self, **kwargs) -> List[str]:
        """Run community discovery with leiden."""
        return []

    async def get_community(self, community_id: str) -> Community:
        """Get community."""
        raise NotImplementedError("Memory graph store does not have community")

    def get_graph_config(self):
        """Get the graph store config."""
        return self._graph_store.get_config()

    def get_vertex_type(self) -> str:
        """Get the vertex type."""
        # raise NotImplementedError("Memory graph store does not have vertex type")
        return ""

    def get_edge_type(self) -> str:
        """Get the edge type."""
        # raise NotImplementedError("Memory graph store does not have edge type")
        return ""

    def get_triplets(self, subj: str) -> List[Tuple[str, str]]:
        """Get triplets."""
        subgraph = self.explore([subj], direct=Direction.OUT, depth=1)
        return [(e.name, e.tid) for e in subgraph.edges()]

    def get_document_vertex(self, doc_name: str) -> Vertex:
        """Get the document vertex in the graph."""
        raise NotImplementedError("Memory graph store does not have document vertex")

    def get_schema(self, refresh: bool = False) -> str:
        """Get the schema of the graph store."""
        return json.dumps(self._graph_store._graph.schema())

    def get_full_graph(self, limit: Optional[int] = None) -> Graph:
        """Get full graph."""
        if not limit:
            return self._graph_store._graph

        subgraph = MemoryGraph()
        for count, edge in enumerate(self._graph_store._graph.edges()):
            if count >= limit:
                break
            subgraph.upsert_vertex(self._graph_store._graph.get_vertex(edge.sid))
            subgraph.upsert_vertex(self._graph_store._graph.get_vertex(edge.tid))
            subgraph.append_edge(edge)
            count += 1
        return subgraph

    def upsert_entities(self, entities: Iterator[Vertex]) -> None:
        """Upsert entities."""
        pass

    def upsert_edge(
        self, edges: Iterator[Edge], edge_type: str, src_type: str, dst_type: str
    ) -> None:
        """Upsert edges."""
        pass

    def upsert_chunks(
        self, chunks: Union[Iterator[Vertex], Iterator[ParagraphChunk]]
    ) -> None:
        """Upsert chunks."""
        pass

    def upsert_documents(
        self, documents: Union[Iterator[Vertex], Iterator[ParagraphChunk]]
    ) -> None:
        """Upsert documents."""
        pass

    def upsert_relations(self, relations: Iterator[Edge]) -> None:
        """Upsert relations."""
        pass

    def upsert_doc_include_chunk(
        self,
        chunk: ParagraphChunk,
    ) -> None:
        """Convert chunk to document include chunk."""
        pass

    def upsert_chunk_include_chunk(
        self,
        chunk: ParagraphChunk,
    ) -> None:
        """Convert chunk to chunk include chunk."""
        pass

    def upsert_chunk_next_chunk(
        self, chunk: ParagraphChunk, next_chunk: ParagraphChunk
    ):
        """Uperst the vertices and the edge in chunk_next_chunk."""
        pass

    def upsert_chunk_include_entity(
        self, chunk: ParagraphChunk, entity: Vertex
    ) -> None:
        """Convert chunk to chunk include entity."""
        pass

    def insert_triplet(self, subj: str, rel: str, obj: str) -> None:
        """Add triplet."""
        self._graph_store._graph.append_edge(Edge(subj, obj, rel))

    def upsert_graph(self, graph: Graph) -> None:
        """Add graph to the graph store.

        Args:
            graph (Graph): The graph to be added.
        """
        for vertex in graph.vertices():
            self._graph_store._graph.upsert_vertex(vertex)

        for edge in graph.edges():
            self._graph_store._graph.append_edge(edge)

    def delete_document(self, chunk_ids: str) -> None:
        """Delete document in the graph."""
        pass

    def delete_triplet(self, sub: str, rel: str, obj: str) -> None:
        """Delete triplet."""
        self._graph_store._graph.del_edges(sub, obj, rel)

    def drop(self):
        """Delete Graph."""
        self._graph_store._graph = None

    def create_graph(self, graph_name: str):
        """Create a graph."""
        pass

    def create_graph_label(
        self,
        graph_elem_type: GraphElemType,
        graph_properties: List[Dict[str, Union[str, bool]]],
    ) -> None:
        """Create a graph label.

        The graph label is used to identify and distinguish different types of nodes
        (vertices) and edges in the graph.
        """
        pass

    def truncate(self):
        """Truncate Graph."""
        self._graph_store._graph.truncate()

    def check_label(self, graph_elem_type: GraphElemType) -> bool:
        """Check if the label exists in the graph.

        Args:
            graph_elem_type (GraphElemType): The type of the graph element.

        Returns:
            True if the label exists in the specified graph element type, otherwise
            False.
        """
        raise NotImplementedError("Memory graph store does not have label")

    def explore(
        self,
        subs: List[str],
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
        search_scope: Optional[
            Literal["knowledge_graph", "document_graph"]
        ] = "knowledge_graph",
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth."""
        return self._graph_store._graph.search(subs, direct, depth, fan, limit)

    def query(self, query: str, **kwargs) -> MemoryGraph:
        """Execute a query on graph."""
        raise NotImplementedError("Memory graph store does not support query")

    async def stream_query(self, query: str, **kwargs) -> AsyncGenerator[Graph, None]:
        """Execute a stream query."""
        raise NotImplementedError("Memory graph store does not support stream query")



"""The summary operator."""

from typing import Any, Optional

from dbgpt.core import LLMClient
from dbgpt.core.awel.flow import IOField, OperatorCategory, Parameter, ViewMetadata
from dbgpt.rag.knowledge.base import Knowledge
from dbgpt.util.i18n_utils import _
from dbgpt_ext.rag.assembler.summary import SummaryAssembler
from dbgpt_ext.rag.operators.assembler import AssemblerOperator


class SummaryAssemblerOperator(AssemblerOperator[Any, Any]):
    """The summary assembler operator."""

    metadata = ViewMetadata(
        label=_("Summary Operator"),
        name="summary_assembler_operator",
        category=OperatorCategory.RAG,
        description=_("The summary assembler operator."),
        inputs=[
            IOField.build_from(
                _("Knowledge"), "knowledge", Knowledge, _("Knowledge datasource")
            )
        ],
        outputs=[
            IOField.build_from(
                _("Document summary"),
                "summary",
                str,
                description="document summary",
            )
        ],
        parameters=[
            Parameter.build_from(
                _("LLM Client"),
                "llm_client",
                LLMClient,
                optional=True,
                default=None,
                description=_("The LLM Client."),
            ),
            Parameter.build_from(
                label=_("Model name"),
                name="model_name",
                type=str,
                optional=True,
                default="gpt-3.5-turbo",
                description=_("LLM model name"),
            ),
            Parameter.build_from(
                label=_("prompt language"),
                name="language",
                type=str,
                optional=True,
                default="en",
                description=_("prompt language"),
            ),
            Parameter.build_from(
                label=_("Max iteration with LLM"),
                name="max_iteration_with_llm",
                type=int,
                optional=True,
                default=5,
                description=_("prompt language"),
            ),
            Parameter.build_from(
                label=_("Concurrency limit with LLM"),
                name="concurrency_limit_with_llm",
                type=int,
                optional=True,
                default=3,
                description=_("The concurrency limit with llm"),
            ),
        ],
        documentation_url="https://github.com/openai/openai-python",
    )

    def __init__(
        self,
        llm_client: Optional[LLMClient],
        model_name: Optional[str] = "gpt-3.5-turbo",
        language: Optional[str] = "en",
        max_iteration_with_llm: Optional[int] = 5,
        concurrency_limit_with_llm: Optional[int] = 3,
        **kwargs,
    ):
        """Create the summary assemble operator.

        Args:
              llm_client: (Optional[LLMClient]) The LLM client.
              model_name: (Optional[str]) The model name.
              language: (Optional[str]) The prompt language.
              max_iteration_with_llm: (Optional[int]) The max iteration with llm.
              concurrency_limit_with_llm: (Optional[int]) The concurrency limit with
                llm.
        """
        super().__init__(**kwargs)
        self._llm_client = llm_client
        self._model_name = model_name
        self._language = language
        self._max_iteration_with_llm = max_iteration_with_llm
        self._concurrency_limit_with_llm = concurrency_limit_with_llm

    async def map(self, knowledge: Knowledge) -> str:
        """Assemble the summary."""
        assembler = SummaryAssembler.load_from_knowledge(
            knowledge=knowledge,
            llm_client=self._llm_client,
            model_name=self._model_name,
            language=self._language,
            max_iteration_with_llm=self._max_iteration_with_llm,
            concurrency_limit_with_llm=self._concurrency_limit_with_llm,
        )
        return await assembler.generate_summary()

    def assemble(self, knowledge: Knowledge) -> Any:
        """Assemble the summary."""
        pass



"""StarRocks connector."""

from dataclasses import dataclass, field
from typing import Any, Iterable, List, Optional, Tuple, Type, cast
from urllib.parse import quote
from urllib.parse import quote_plus as urlquote

from sqlalchemy import text

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.rdbms.base import RDBMSConnector, RDBMSDatasourceParameters
from dbgpt.util.i18n_utils import _

from .dialect.starrocks.sqlalchemy import *  # noqa


@auto_register_resource(
    label=_("StarRocks datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("An Open-Source, High-Performance Analytical Database."),
)
@dataclass
class StarRocksParameters(RDBMSDatasourceParameters):
    """StarRocks connection parameters."""

    __type__ = "starrocks"

    driver: str = field(
        default="starrocks",
        metadata={
            "help": _("Driver name for starrocks, default is starrocks."),
        },
    )

    def create_connector(self) -> "StarRocksConnector":
        """Create StarRocks connector."""
        return StarRocksConnector.from_parameters(self)


class StarRocksConnector(RDBMSConnector):
    """StarRocks connector."""

    driver = "starrocks"
    db_type = "starrocks"
    db_dialect = "starrocks"

    @classmethod
    def param_class(cls) -> Type[StarRocksParameters]:
        """Return the parameter class."""
        return StarRocksParameters

    @classmethod
    def from_uri_db(
        cls: Type["StarRocksConnector"],
        host: str,
        port: int,
        user: str,
        pwd: str,
        db_name: str,
        engine_args: Optional[dict] = None,
        **kwargs: Any,
    ) -> "StarRocksConnector":
        """Create a new StarRocksConnector from host, port, user, pwd, db_name."""
        db_url: str = (
            f"{cls.driver}://{quote(user)}:{urlquote(pwd)}@{host}:{str(port)}/{db_name}"
        )
        return cast(StarRocksConnector, cls.from_uri(db_url, engine_args, **kwargs))

    def _sync_tables_from_db(self) -> Iterable[str]:
        db_name = self.get_current_db_name()
        with self.session_scope() as session:
            table_results = session.execute(
                text(
                    "SELECT TABLE_NAME FROM information_schema.tables where "
                    f'TABLE_SCHEMA="{db_name}"'
                )
            )
            # view_results = session.execute(text(f'SELECT TABLE_NAME from
            # information_schema.materialized_views where TABLE_SCHEMA="{db_name}"'))
            table_results = set(row[0] for row in table_results)  # noqa: C401
            # view_results = set(row[0] for row in view_results)
            self._all_tables = table_results
            self._metadata.reflect(bind=self._engine)
            return self._all_tables

    def get_grants(self):
        """Get grants."""
        with self.session_scope() as session:
            cursor = session.execute(text("SHOW GRANTS"))
            grants = cursor.fetchall()
            if len(grants) == 0:
                return []
            if len(grants[0]) == 2:
                grants_list = [x[1] for x in grants]
            else:
                grants_list = [x[2] for x in grants]
            return grants_list

    def _get_current_version(self):
        """Get database current version."""
        with self.session_scope() as session:
            return int(session.execute(text("select current_version()")).scalar())

    def get_collation(self):
        """Get collation."""
        # StarRocks 排序是表级别的
        return None

    def get_users(self):
        """Get user info."""
        return []

    def get_fields(self, table_name, db_name="database()") -> List[Tuple]:
        """Get column fields about specified table."""
        with self.session_scope() as session:
            if db_name != "database()":
                db_name = f'"{db_name}"'
            cursor = session.execute(
                text(
                    "select COLUMN_NAME, COLUMN_TYPE, COLUMN_DEFAULT, IS_NULLABLE, "
                    "COLUMN_COMMENT from information_schema.columns where "
                    f'TABLE_NAME="{table_name}" and TABLE_SCHEMA = {db_name}'
                )
            )
            fields = cursor.fetchall()
            return [
                (field[0], field[1], field[2], field[3], field[4]) for field in fields
            ]

    def get_charset(self):
        """Get character_set."""
        return "utf-8"

    def get_show_create_table(self, table_name: str):
        """Get show create table."""
        # cur = self.session.execute(
        #     text(
        #         f"""show create table {table_name}"""
        #     )
        # )
        # rows = cur.fetchone()
        # create_sql = rows[0]

        # return create_sql
        # Here is the table description, returning the create table statement will
        # cause the token to be too long and fail
        with self.session_scope() as session:
            cur = session.execute(
                text(
                    "SELECT TABLE_COMMENT FROM information_schema.tables where "
                    f'TABLE_NAME="{table_name}" and TABLE_SCHEMA=database()'
                )
            )
            table = cur.fetchone()
            if table:
                return str(table[0])
            else:
                return ""

    def get_table_comments(self, db_name=None):
        """Get table comments."""
        if not db_name:
            db_name = self.get_current_db_name()
        with self.session_scope() as session:
            cur = session.execute(
                text(
                    "SELECT TABLE_NAME,TABLE_COMMENT FROM information_schema.tables "
                    f'where TABLE_SCHEMA="{db_name}"'
                )
            )
            tables = cur.fetchall()
            return [(table[0], table[1]) for table in tables]

    def get_database_names(self):
        """Get database names."""
        with self.session_scope() as session:
            cursor = session.execute(text("SHOW DATABASES;"))
            results = cursor.fetchall()
            return [
                d[0]
                for d in results
                if d[0] not in ["information_schema", "sys", "_statistics_", "dataease"]
            ]

    def get_current_db_name(self) -> str:
        """Get current database name."""
        with self.session_scope() as session:
            return session.execute(text("select database()")).scalar()

    def table_simple_info(self):
        """Get table simple info."""
        _sql = """
          SELECT concat(TABLE_NAME,"(",group_concat(COLUMN_NAME,","),");")
           FROM information_schema.columns where TABLE_SCHEMA=database()
            GROUP BY TABLE_NAME
        """
        with self.session_scope() as session:
            cursor = session.execute(text(_sql))
            results = cursor.fetchall()
            return [x[0] for x in results]

    def get_indexes(self, table_name):
        """Get table indexes about specified table."""
        with self.session_scope() as session:
            cursor = session.execute(text(f"SHOW INDEX FROM {table_name}"))
            indexes = cursor.fetchall()
            return [(index[2], index[4]) for index in indexes]



"""Text Based Graph Retriever."""

import json
import logging
from typing import Dict, List, Tuple, Union

from dbgpt.storage.graph_store.graph import Graph, MemoryGraph
from dbgpt_ext.rag.retriever.graph_retriever.base import GraphRetrieverBase

logger = logging.getLogger(__name__)


class TextBasedGraphRetriever(GraphRetrieverBase):
    """Text Based Graph Retriever class."""

    def __init__(
        self,
        graph_store_adapter,
        triplet_topk,
        intent_interpreter,
        text2gql,
    ):
        """Initialize Text Based Graph Retriever."""
        self._graph_store_adapter = graph_store_adapter
        self._triplet_topk = triplet_topk
        self._intent_interpreter = intent_interpreter
        self._text2gql = text2gql

    async def retrieve(self, text: str) -> Tuple[Graph, str]:
        """Retrieve from triplets graph with text2gql."""
        intention: Dict[
            str, Union[str, List[str]]
        ] = await self._intent_interpreter.translate(text)
        schema = json.dumps(
            json.loads(self._graph_store_adapter.get_schema()), indent=4
        )
        intention["schema"] = schema
        translation: Dict[str, str] = await self._text2gql.translate(
            json.dumps(intention)
        )
        text2gql_query = translation.get("query", "")
        if "LIMIT" not in text2gql_query:
            text2gql_query += f" LIMIT {self._triplet_topk}"
        try:
            subgraph = self._graph_store_adapter.query(query=text2gql_query)
            logger.info(f"Query executed successfully: {text2gql_query}")
        except Exception as e:
            text2gql_query = ""
            subgraph = MemoryGraph()
            logger.error(f"Failed to execute query: {text2gql_query}\n{e}")

        return subgraph, text2gql_query



"""Neo4j Connector."""

from dataclasses import dataclass, field
from typing import Dict, Generator, Iterator, List, Type, cast

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.base import BaseConnector
from dbgpt.datasource.parameter import BaseDatasourceParameters
from dbgpt.util.i18n_utils import _


@auto_register_resource(
    label=_("Neo4j datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_(
        "Neo4j is a highly scalable native graph database, purpose-built to leverage "
        "data relationships."
    ),
    icon="icons/neo4j.png",
)
@dataclass
class Neo4jParameters(BaseDatasourceParameters):
    """Neo4j connection parameters."""

    __type__ = "neo4j"

    host: str = field(metadata={"help": _("Neo4j server host")})
    user: str = field(metadata={"help": _("Neo4j server user")})
    password: str = field(
        default="${env:DBGPT_DB_PASSWORD}",
        metadata={
            "help": _(
                "Database password, you can write your password directly, of course, "
                "you can also use environment variables, such as "
                "${env:DBGPT_DB_PASSWORD}"
            ),
            "tags": "privacy",
        },
    )
    port: int = field(
        default=7687, metadata={"help": _("Neo4j server port, default 7687")}
    )
    database: str = field(
        default="neo4j", metadata={"help": _("Database name, default 'neo4j'")}
    )

    def create_connector(self) -> "BaseConnector":
        """Create Neo4j connector."""
        return Neo4jConnector.from_parameters(self)

    def db_url(self, ssl=False, charset=None):
        """Get the database URL."""
        raise NotImplementedError("Neo4j does not support db_url")


class Neo4jConnector(BaseConnector):
    """Neo4j connector."""

    db_type: str = "neo4j"
    driver: str = "bolt"
    dialect: str = "cypher"

    def __init__(self, driver, database):
        """Initialize the connector with a Neo4j driver."""
        self._driver = driver
        self._database = database
        self._schema = None
        self._is_closed = False
        self._graph = database  # Alias for compatibility

    @classmethod
    def param_class(cls) -> Type[Neo4jParameters]:
        """Return the parameter class."""
        return Neo4jParameters

    @classmethod
    def from_parameters(cls, parameters: Neo4jParameters) -> "Neo4jConnector":
        """Create a new Neo4jConnector from parameters."""
        return cls.from_uri_db(
            parameters.host,
            parameters.port,
            parameters.user,
            parameters.password,
            parameters.database,
        )

    @classmethod
    def from_uri_db(
        cls, host: str, port: int, user: str, pwd: str, db_name: str, **kwargs
    ) -> "Neo4jConnector":
        """Create a new Neo4jConnector from host, port, user, pwd, db_name.

        Args:
            host: Neo4j server host
            port: Neo4j server port
            user: Neo4j username
            pwd: Neo4j password
            db_name: Database name
            **kwargs: Additional arguments (e.g., schema) are ignored for Neo4j
        """
        try:
            from neo4j import GraphDatabase

            db_url = f"{cls.driver}://{host}:{str(port)}"
            driver = GraphDatabase.driver(db_url, auth=(user, pwd))
            driver.verify_connectivity()
            return cast(Neo4jConnector, cls(driver=driver, database=db_name))

        except ImportError as err:
            raise ImportError(
                "neo4j package is not installed, please install it with "
                "`pip install neo4j`"
            ) from err

    def create_graph(self, graph_name: str) -> bool:
        """Create a graph in Neo4j.

        Note: In Neo4j, databases are created at the system level.
        For knowledge graphs, we don't need to explicitly create a graph structure.
        This method is here for compatibility with TuGraph adapter.

        Returns:
            bool: True (graph always exists in Neo4j)
        """
        # Neo4j databases exist once created at system level
        # For knowledge graph usage, we just need to ensure the database exists
        return True

    def is_exist(self, graph_name: str) -> bool:
        """Check if a database exists in Neo4j.

        Args:
            graph_name: Database name to check

        Returns:
            bool: True if database exists
        """
        try:
            databases = self.get_database_names()
            return graph_name in databases
        except Exception:
            # If we can connect and run queries, the database exists
            return True

    def delete_graph(self, graph_name: str) -> None:
        """Delete all data from the current database.

        Note: This deletes all nodes and relationships, not the database itself.
        """
        with self._driver.session(database=self._database) as session:
            # Delete all relationships first
            session.run("MATCH ()-[r]->() DELETE r")
            # Then delete all nodes
            session.run("MATCH (n) DELETE n")

    def get_system_info(self) -> Dict:
        """Get Neo4j system information."""
        system_info = {}
        try:
            with self._driver.session(database="system") as session:
                # Get Neo4j version
                result = session.run("CALL dbms.components() YIELD versions")
                for record in result:
                    versions = record["versions"]
                    if versions:
                        system_info["neo4j_version"] = versions[0]
                        # For compatibility with TuGraph adapter
                        system_info["lgraph_version"] = versions[0]
        except Exception:
            # Fallback for older Neo4j versions
            system_info["neo4j_version"] = "unknown"
            system_info["lgraph_version"] = "5.0.0"  # Default compatible version

        return system_info

    def get_table_names(self) -> Iterator[str]:
        """Get all table names from Neo4j (node labels and relationship types)."""
        with self._driver.session(database=self._database) as session:
            # Get node labels
            result = session.run("CALL db.labels()")
            node_labels = [record["label"] + "_node" for record in result]

            # Get relationship types
            result = session.run("CALL db.relationshipTypes()")
            rel_types = [
                record["relationshipType"] + "_relationship" for record in result
            ]

            return iter(node_labels + rel_types)

    def get_columns(self, table_name: str, table_type: str = "node") -> List[Dict]:
        """Retrieve the properties for a specified node label or relationship type.

        Args:
            table_name (str): node label or relationship type name
            table_type (str): table type ('node' or 'relationship')

        Returns:
            columns: List[Dict], which contains name: str, type: str,
                default_expression: str, is_in_primary_key: bool, comment: str
        """
        with self._driver.session(database=self._database) as session:
            # Check if label/relationship exists first to avoid warnings
            if table_type == "node":
                # Remove _node suffix if present
                label_name = table_name.replace("_node", "")
                check_result = session.run("CALL db.labels()")
                existing_labels = [record["label"] for record in check_result]
                if label_name not in existing_labels:
                    return []

                # Get properties for node label
                query = f"""
                MATCH (n:`{label_name}`)
                WITH n LIMIT 100
                UNWIND keys(n) AS key
                RETURN DISTINCT key AS property
                """
            else:
                # Remove _relationship suffix if present
                rel_name = table_name.replace("_relationship", "")
                check_result = session.run("CALL db.relationshipTypes()")
                existing_rels = [record["relationshipType"] for record in check_result]
                if rel_name not in existing_rels:
                    return []

                # Get properties for relationship type
                query = f"""
                MATCH ()-[r:`{rel_name}`]->()
                WITH r LIMIT 100
                UNWIND keys(r) AS key
                RETURN DISTINCT key AS property
                """

            result = session.run(query)
            properties = []
            for record in result:
                prop_dict = {
                    "name": record["property"],
                    "type": "string",  # Neo4j is schema-less, type inference is complex
                    "default_expression": "",
                    "is_in_primary_key": False,
                    "comment": record["property"],
                }
                properties.append(prop_dict)
            return properties

    def get_indexes(self, table_name: str, table_type: str = "node") -> List[Dict]:
        """Get indexes for a specified node label.

        Args:
            table_name (str): node label name
            table_type (str): 'node' | 'relationship'

        Returns:
            List[Dict]: eg:[{'name': 'idx_key', 'column_names': ['id']}]
        """
        with self._driver.session(database=self._database) as session:
            result = session.run("SHOW INDEXES")
            indexes = []
            for record in result:
                # Check if this index applies to the given label
                labels_or_types = record.get("labelsOrTypes") or []
                if table_name in labels_or_types:
                    index_dict = {
                        "name": record.get("name", ""),
                        "column_names": record.get("properties") or [],
                    }
                    indexes.append(index_dict)
            return indexes

    def get_grants(self):
        """Get grants."""
        return []

    def get_collation(self):
        """Get collation."""
        return "UTF-8"

    def get_charset(self):
        """Get character_set of current database."""
        return "UTF-8"

    def table_simple_info(self):
        """Get table simple info."""
        return []

    def close(self):
        """Close the Neo4j driver."""
        if self._is_closed:
            return
        self._driver.close()
        self._is_closed = True

    def run(self, query: str, fetch: str = "all", **params) -> List:
        """Run Cypher query.

        Args:
            query (str): Cypher query string
            fetch (str): fetch type ('all' or 'one')
            **params: Query parameters

        Returns:
            List: query results
        """
        with self._driver.session(database=self._database) as session:
            try:
                result = session.run(query, **params)
                return list(result)
            except Exception as e:
                raise Exception(f"Query execution failed: {e}\nQuery: {query}") from e

    def run_stream(self, query: str) -> Generator:
        """Run Cypher query with streaming results.

        Args:
            query (str): Cypher query string

        Yields:
            Query results one by one
        """
        with self._driver.session(database=self._database) as session:
            result = session.run(query)
            yield from result

    def get_database_names(self) -> List[str]:
        """Get all database names.

        Returns:
            List[str]: list of database names
        """
        with self._driver.session(database="system") as session:
            result = session.run("SHOW DATABASES")
            return [record["name"] for record in result]

    @classmethod
    def is_graph_type(cls) -> bool:
        """Return whether the connector is a graph database connector."""
        return True



"""Embedding Assembler."""

from concurrent.futures import ThreadPoolExecutor
from typing import Any, List, Optional

from dbgpt.core import Chunk, Embeddings
from dbgpt.rag.knowledge.base import Knowledge
from dbgpt.rag.retriever import BaseRetriever, RetrieverStrategy
from dbgpt.rag.retriever.embedding import EmbeddingRetriever
from dbgpt.storage.base import IndexStoreBase
from dbgpt.util.executor_utils import blocking_func_to_async

from ..assembler.base import BaseAssembler
from ..chunk_manager import ChunkParameters


class EmbeddingAssembler(BaseAssembler):
    """Embedding Assembler.

    Example:
    .. code-block:: python

        from dbgpt_ext.rag.assembler import EmbeddingAssembler

        pdf_path = "path/to/document.pdf"
        knowledge = KnowledgeFactory.from_file_path(pdf_path)
        assembler = EmbeddingAssembler.load_from_knowledge(
            knowledge=knowledge,
            embedding_model="text2vec",
        )
    """

    def __init__(
        self,
        knowledge: Knowledge,
        index_store: IndexStoreBase,
        chunk_parameters: Optional[ChunkParameters] = None,
        retrieve_strategy: Optional[RetrieverStrategy] = RetrieverStrategy.EMBEDDING,
        **kwargs: Any,
    ) -> None:
        """Initialize with Embedding Assembler arguments.

        Args:
            knowledge: (Knowledge) Knowledge datasource.
            index_store: (IndexStoreBase) IndexStoreBase to use.
            chunk_parameters: (Optional[ChunkParameters]) ChunkManager to use for
                chunking.
            keyword_store: (Optional[IndexStoreBase]) IndexStoreBase to use.
            embedding_model: (Optional[str]) Embedding model to use.
            embeddings: (Optional[Embeddings]) Embeddings to use.
        """
        if knowledge is None:
            raise ValueError("knowledge datasource must be provided.")
        self._index_store = index_store
        self._retrieve_strategy = retrieve_strategy

        super().__init__(
            knowledge=knowledge,
            chunk_parameters=chunk_parameters,
            **kwargs,
        )

    @classmethod
    def load_from_knowledge(
        cls,
        knowledge: Knowledge,
        index_store: IndexStoreBase,
        chunk_parameters: Optional[ChunkParameters] = None,
        embedding_model: Optional[str] = None,
        embeddings: Optional[Embeddings] = None,
        retrieve_strategy: Optional[RetrieverStrategy] = RetrieverStrategy.EMBEDDING,
    ) -> "EmbeddingAssembler":
        """Load document embedding into vector store from path.

        Args:
            knowledge: (Knowledge) Knowledge datasource.
            index_store: (IndexStoreBase) IndexStoreBase to use.
            chunk_parameters: (Optional[ChunkParameters]) ChunkManager to use for
                chunking.
            embedding_model: (Optional[str]) Embedding model to use.
            embeddings: (Optional[Embeddings]) Embeddings to use.
            retrieve_strategy: (Optional[RetrieverStrategy]) Retriever strategy.

        Returns:
             EmbeddingAssembler
        """
        return cls(
            knowledge=knowledge,
            index_store=index_store,
            chunk_parameters=chunk_parameters,
            embedding_model=embedding_model,
            embeddings=embeddings,
            retrieve_strategy=retrieve_strategy,
        )

    @classmethod
    async def aload_from_knowledge(
        cls,
        knowledge: Knowledge,
        index_store: IndexStoreBase,
        chunk_parameters: Optional[ChunkParameters] = None,
        executor: Optional[ThreadPoolExecutor] = None,
        retrieve_strategy: Optional[RetrieverStrategy] = RetrieverStrategy.EMBEDDING,
    ) -> "EmbeddingAssembler":
        """Load document embedding into vector store from path.

        Args:
            knowledge: (Knowledge) Knowledge datasource.
            chunk_parameters: (Optional[ChunkParameters]) ChunkManager to use for
                chunking.
            index_store: (IndexStoreBase) Index store to use.
            executor: (Optional[ThreadPoolExecutor) ThreadPoolExecutor to use.
            retrieve_strategy: (Optional[RetrieverStrategy]) Retriever strategy.

        Returns:
             EmbeddingAssembler
        """
        executor = executor or ThreadPoolExecutor()
        return await blocking_func_to_async(
            executor,
            cls,
            knowledge,
            index_store,
            chunk_parameters,
            retrieve_strategy,
        )

    def persist(self, **kwargs) -> List[str]:
        """Persist chunks into store.

        Returns:
            List[str]: List of chunk ids.
        """
        max_chunks_once_load = kwargs.get("max_chunks_once_load")
        max_threads = kwargs.get("max_threads")
        return self._index_store.load_document_with_limit(
            self._chunks, max_chunks_once_load, max_threads
        )

    async def apersist(self, **kwargs) -> List[str]:
        """Persist chunks into store.

        Returns:
            List[str]: List of chunk ids.
        """
        # persist chunks into vector store
        max_chunks_once_load = kwargs.get("max_chunks_once_load")
        max_threads = kwargs.get("max_threads")
        file_id = kwargs.get("file_id", None)
        return await self._index_store.aload_document_with_limit(
            self._chunks, max_chunks_once_load, max_threads, file_id
        )

    def _extract_info(self, chunks) -> List[Chunk]:
        """Extract info from chunks."""
        return []

    def as_retriever(self, top_k: int = 4, **kwargs) -> BaseRetriever:
        """Create a retriever.

        Args:
            top_k(int): default 4.

        Returns:
            EmbeddingRetriever
        """
        return EmbeddingRetriever(
            top_k=top_k,
            index_store=self._index_store,
            retrieve_strategy=self._retrieve_strategy,
        )



"""Doris connector."""

import weakref
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, cast
from urllib.parse import quote
from urllib.parse import quote_plus as urlquote

from sqlalchemy import MetaData, inspect, text
from sqlalchemy.orm import scoped_session, sessionmaker

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.rdbms.base import RDBMSConnector, RDBMSDatasourceParameters
from dbgpt.util.i18n_utils import _


@auto_register_resource(
    label=_("Apache Doris datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("A new-generation open-source real-time data warehouse."),
)
@dataclass
class DorisParameters(RDBMSDatasourceParameters):
    """Doris connection parameters.

    Doris has a same protocol with MySQL, so we suggest to use MySQL connector to
    connect Doris.
    """

    __type__ = "doris"
    driver: str = field(
        default="mysql+pymysql",
        metadata={
            "help": _(
                "Driver name for Doris, default is mysql+pymysql (MySQL compatible)."
            ),
        },
    )

    def create_connector(self) -> "DorisConnector":
        """Create doris connector"""
        return DorisConnector.from_parameters(self)


class DorisConnector(RDBMSConnector):
    """Doris connector."""

    driver = "mysql+pymysql"
    db_type = "doris"
    db_dialect = "mysql"

    def __init__(
        self,
        engine,
        schema: Optional[str] = None,
        metadata: Optional[MetaData] = None,
        ignore_tables: Optional[List[str]] = None,
        include_tables: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
        indexes_in_table_info: bool = False,
        custom_table_info: Optional[Dict[str, str]] = None,
        view_support: bool = False,
    ):
        """Initialize Doris connector without triggering reflection.

        Override parent __init__ to avoid automatic metadata.reflect() call
        which causes issues with Doris data type parsing.
        """
        # Initialize basic attributes (copied from parent but without reflect)
        self._is_closed = False
        self._engine = engine
        self._schema = schema
        if include_tables and ignore_tables:
            raise ValueError("Cannot specify both include_tables and ignore_tables")

        if not custom_table_info:
            custom_table_info = {}

        self._inspector = inspect(engine)
        session_factory = sessionmaker(bind=engine)
        Session_Manages = scoped_session(session_factory)
        self._db_sessions = Session_Manages
        self._sessions = weakref.WeakSet()

        self.view_support = view_support
        self._usable_tables = set()
        self._include_tables = set()
        self._ignore_tables = set()
        self._custom_table_info = custom_table_info
        self._sample_rows_in_table_info = sample_rows_in_table_info
        self._indexes_in_table_info = indexes_in_table_info

        # NOT call reflect() to avoid Doris type parsing issues
        # self._metadata = metadata or MetaData()
        # self._metadata.reflect(bind=self._engine)

        self._all_tables = set(self._sync_tables_from_db())

    @classmethod
    def param_class(cls) -> Type[DorisParameters]:
        """Return the parameter class."""
        return DorisParameters

    @classmethod
    def from_uri_db(
        cls,
        host: str,
        port: int,
        user: str,
        pwd: str,
        db_name: str,
        engine_args: Optional[dict] = None,
        **kwargs: Any,
    ) -> "DorisConnector":
        """Create a new DorisConnector from host, port, user, pwd, db_name."""
        db_url: str = (
            f"{cls.driver}://{quote(user)}:{urlquote(pwd)}@{host}:{str(port)}/{db_name}"
        )
        return cast(DorisConnector, cls.from_uri(db_url, engine_args, **kwargs))

    def _sync_tables_from_db(self) -> Iterable[str]:
        """Sync tables from db."""
        with self.session_scope() as session:
            table_results = session.execute(
                text(
                    "SELECT TABLE_NAME FROM information_schema.tables where "
                    "TABLE_SCHEMA=database()"
                )
            )
            table_results = set(row[0] for row in table_results)  # noqa: C401
            self._all_tables = table_results
            return self._all_tables

    def get_grants(self):
        """Get grants."""
        with self.session_scope() as session:
            cursor = session.execute(text("SHOW GRANTS"))
            grants = cursor.fetchall()
            if len(grants) == 0:
                return []
            if len(grants[0]) == 2:
                grants_list = [x[1] for x in grants]
            else:
                grants_list = [x[2] for x in grants]
            return grants_list

    def _get_current_version(self):
        """Get database current version."""
        with self.session_scope() as session:
            return int(session.execute(text("select current_version()")).scalar())

    def get_collation(self):
        """Get collation.

        ref `SHOW COLLATION <https://doris.apache.org/zh-CN/docs/dev/sql-manual/
        sql-reference/Show-Statements/SHOW-COLLATION/>`_

        """
        with self.session_scope() as session:
            cursor = session.execute(text("SHOW COLLATION"))
            results = cursor.fetchall()
            return "" if not results else results[0][0]

    def get_users(self):
        """Get user info."""
        return []

    def get_columns(self, table_name: str) -> List[Dict]:
        """Get columns.

        Args:
            table_name (str): str
        Returns:
            columns: List[Dict], which contains name: str, type: str,
                default_expression: str, is_in_primary_key: bool, comment: str
                eg:[{'name': 'id', 'type': 'UInt64', 'default_expression': '',
                'is_in_primary_key': True, 'comment': 'id'}, ...]
        """
        fields = self.get_fields(table_name)
        return [
            {
                "name": field[0],
                "type": field[1],
                "default": field[2],
                "nullable": field[3],
                "comment": field[4],
            }
            for field in fields
        ]

    def get_fields(self, table_name, db_name=None) -> List[Tuple]:
        """Get column fields about specified table."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    "select COLUMN_NAME, COLUMN_TYPE, COLUMN_DEFAULT, IS_NULLABLE, "
                    "COLUMN_COMMENT from information_schema.columns "
                    f'where TABLE_NAME="{table_name}" and TABLE_SCHEMA=database()'
                )
            )
            fields = cursor.fetchall()
            return [
                (field[0], field[1], field[2], field[3], field[4]) for field in fields
            ]

    def get_charset(self) -> str:
        """Get character_set."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    """
                    SELECT DEFAULT_CHARACTER_SET_NAME 
                    FROM information_schema.SCHEMATA 
                    where SCHEMA_NAME=database() 
                    """
                )
            )
            ans = cursor.fetchall()
            if ans:
                return ans[0][0]
            return ""

    def get_show_create_table(self, table_name) -> str:
        """Get show create table."""
        # cur = self.get_session().execute(
        #     text(
        #         f"""show create table {table_name}"""
        #     )
        # )
        # rows = cur.fetchone()
        # create_sql = rows[1]
        # return create_sql
        # Here is the table description, returning the create table statement will
        with self.session_scope() as session:
            cur = session.execute(
                text(
                    f"SELECT TABLE_COMMENT "
                    f"FROM information_schema.tables "
                    f'where TABLE_NAME="{table_name}" and TABLE_SCHEMA=database()'
                )
            )
            table = cur.fetchone()
            if table:
                return str(table[0])
            else:
                return ""

    def get_table_comments(self, db_name=None):
        """Get table comments."""
        db_name = "database()" if not db_name else f"'{db_name}'"
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    f"SELECT TABLE_NAME,TABLE_COMMENT "
                    f"FROM information_schema.tables "
                    f"where TABLE_SCHEMA={db_name}"
                )
            )
            tables = cursor.fetchall()
            return [(table[0], table[1]) for table in tables]

    def get_table_comment(self, table_name=None):
        """Get table comment."""
        cursor = self.get_session().execute(
            text(
                f"SELECT TABLE_COMMENT "
                f"FROM information_schema.tables "
                f"where TABLE_NAME={table_name} and TABLE_SCHEMA=database()"
            )
        )
        table_comment = cursor.fetchone()
        if table_comment:
            return {"text": str(table_comment[0])}
        else:
            return {"text": None}

    def get_database_names(self):
        """Get database names."""
        with self.session_scope() as session:
            cursor = session.execute(text("SHOW DATABASES"))
            results = cursor.fetchall()
            return [
                d[0]
                for d in results
                if d[0]
                not in [
                    "information_schema",
                    "sys",
                    "_statistics_",
                    "mysql",
                    "__internal_schema",
                    "doris_audit_db__",
                ]
            ]

    def get_current_db_name(self) -> str:
        """Get current database name."""
        with self.session_scope() as session:
            return session.execute(text("select database()")).scalar()

    def table_simple_info(self):
        """Get table simple info."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    "SELECT concat(TABLE_NAME,'(',group_concat(COLUMN_NAME,','),');') "
                    "FROM information_schema.columns "
                    "where TABLE_SCHEMA=database() "
                    "GROUP BY TABLE_NAME"
                )
            )
            results = cursor.fetchall()
            return [x[0] for x in results]

    def get_indexes(self, table_name):
        """Get table indexes about specified table."""
        with self.session_scope() as session:
            cursor = session.execute(text(f"SHOW INDEX FROM {table_name}"))
            indexes = cursor.fetchall()
            return [(index[2], index[4]) for index in indexes]

    def get_table_info(self, table_names: Optional[List[str]] = None) -> str:
        """Get information about specified tables.

        Override parent method to avoid dependency on metadata.reflect()
        which causes issues with Doris data type parsing.
        Uses direct SQL queries to get table information.
        """
        all_table_names = list(self.get_usable_table_names())
        if table_names is not None:
            missing_tables = set(table_names).difference(all_table_names)
            if missing_tables:
                raise ValueError(f"table_names {missing_tables} not found in database")
            all_table_names = table_names

        if not all_table_names:
            return ""

        tables = []
        for table_name in all_table_names:
            if self._custom_table_info and table_name in self._custom_table_info:
                tables.append(self._custom_table_info[table_name])
                continue

            # Build table info using direct SQL queries
            table_info = self._build_table_info_for_doris(table_name)
            tables.append(table_info)

        return "\n\n".join(tables)

    def _build_table_info_for_doris(self, table_name: str) -> str:
        """Build table information for Doris using direct SQL queries."""
        try:
            with self.session_scope() as session:
                # Get table structure information
                cursor = session.execute(
                    text(
                        "SELECT COLUMN_NAME, COLUMN_TYPE, IS_NULLABLE, "
                        "COLUMN_DEFAULT, COLUMN_COMMENT "
                        "FROM information_schema.columns "
                        f'WHERE TABLE_NAME="{table_name}" AND TABLE_SCHEMA=database() '
                        "ORDER BY ORDINAL_POSITION"
                    )
                )
                columns = cursor.fetchall()

                if not columns:
                    return f"-- Table {table_name} not found"

                # Build CREATE TABLE statement
                table_info = f"CREATE TABLE {table_name} (\n"
                column_definitions = []

                for col in columns:
                    col_name, col_type, is_nullable, col_default, col_comment = col
                    col_def = f"  `{col_name}` {col_type}"

                    if is_nullable == "NO":
                        col_def += " NOT NULL"

                    if col_default is not None:
                        col_def += f" DEFAULT {col_default}"

                    if col_comment:
                        col_def += f" COMMENT '{col_comment}'"

                    column_definitions.append(col_def)

                table_info += ",\n".join(column_definitions)
                table_info += "\n)"

                # Get table comment if available
                try:
                    comment_cursor = session.execute(
                        text(
                            "SELECT TABLE_COMMENT FROM information_schema.tables "
                            f'WHERE TABLE_NAME="{table_name}"'
                            f" AND TABLE_SCHEMA=database()"
                        )
                    )
                    table_comment = comment_cursor.fetchone()
                    if table_comment and table_comment[0]:
                        table_info += f" COMMENT='{table_comment[0]}'"
                except Exception:
                    pass  # Ignore comment retrieval errors

                # Add sample rows if configured
                if self._sample_rows_in_table_info > 0:
                    table_info += self._get_sample_rows_for_doris(table_name)

                # Add index information if configured
                if self._indexes_in_table_info:
                    table_info += self._get_indexes_info_for_doris(table_name)

                return table_info

        except Exception as e:
            return f"-- Error getting info for table {table_name}: {str(e)}"

    def _get_sample_rows_for_doris(self, table_name: str) -> str:
        """Get sample rows for Doris table."""
        try:
            with self.session_scope() as session:
                cursor = session.execute(
                    text(
                        f"SELECT * FROM {table_name} LIMIT "
                        f"{self._sample_rows_in_table_info}"
                    )
                )
                rows = cursor.fetchall()

                if not rows:
                    return ""

                # Get column names
                column_names = list(cursor.keys())
                columns_str = "\t".join(column_names)

                # Format sample rows
                sample_rows_str = "\n".join(
                    [
                        "\t".join(
                            [
                                str(val)[:100] if val is not None else "NULL"
                                for val in row
                            ]
                        )
                        for row in rows
                    ]
                )

                return (
                    f"\n\n/*\n{self._sample_rows_in_table_info} rows from "
                    f"{table_name} table:\n{columns_str}\n{sample_rows_str}\n*/"
                )

        except Exception:
            return f"\n\n/*\nError getting sample rows for table {table_name}\n*/"

    def _get_indexes_info_for_doris(self, table_name: str) -> str:
        """Get index information for Doris table."""
        try:
            indexes = self.get_indexes(table_name)
            if not indexes:
                return f"\n\n/*\nTable Indexes for {table_name}:\nNo indexes found\n*/"

            indexes_str = "\n".join(
                [f"Index: {idx[0]}, Column: {idx[1]}" for idx in indexes]
            )
            return f"\n\n/*\nTable Indexes for {table_name}:\n{indexes_str}\n*/"

        except Exception:
            return f"\n\n/*\nError getting indexes for table {table_name}\n*/"



"""Assembler Module For RAG.

The Assembler is a module that is responsible for assembling the knowledge.
"""

from .base import BaseAssembler  # noqa: F401
from .db_schema import DBSchemaAssembler  # noqa: F401
from .embedding import EmbeddingAssembler  # noqa: F401
from .summary import SummaryAssembler  # noqa: F401

__all__ = [
    "BaseAssembler",
    "DBSchemaAssembler",
    "EmbeddingAssembler",
    "SummaryAssembler",
]



"""OB Dialect support."""

import re

from sqlalchemy import util
from sqlalchemy.dialects import registry
from sqlalchemy.dialects.mysql import pymysql
from sqlalchemy.dialects.mysql.reflection import MySQLTableDefinitionParser, _re_compile


class OceanBaseTableDefinitionParser(MySQLTableDefinitionParser):
    """OceanBase table definition parser."""

    def __init__(self, dialect, preparer, *, default_schema=None):
        """Initialize OceanBaseTableDefinitionParser."""
        MySQLTableDefinitionParser.__init__(self, dialect, preparer)
        self.default_schema = default_schema

    def _prep_regexes(self):
        super()._prep_regexes()

        _final = self.preparer.final_quote
        quotes = dict(
            zip(
                ("iq", "fq", "esc_fq"),
                [
                    re.escape(s)
                    for s in (
                        self.preparer.initial_quote,
                        _final,
                        self.preparer._escape_identifier(_final),
                    )
                ],
            )
        )

        self._re_key = _re_compile(
            r"  "
            r"(?:(SPATIAL|VECTOR|(?P<type>\S+)) )?KEY"
            # r"(?:(?P<type>\S+) )?KEY"
            r"(?: +{iq}(?P<name>(?:{esc_fq}|[^{fq}])+){fq})?"
            r"(?: +USING +(?P<using_pre>\S+))?"
            r" +\((?P<columns>.+?)\)"
            r"(?: +USING +(?P<using_post>\S+))?"
            r"(?: +(KEY_)?BLOCK_SIZE *[ =]? *(?P<keyblock>\S+) *(LOCAL)?)?"
            r"(?: +WITH PARSER +(?P<parser>\S+))?"
            r"(?: +COMMENT +(?P<comment>(\x27\x27|\x27([^\x27])*?\x27)+))?"
            r"(?: +/\*(?P<version_sql>.+)\*/ *)?"
            r",?$".format(iq=quotes["iq"], esc_fq=quotes["esc_fq"], fq=quotes["fq"])
        )

        kw = quotes.copy()
        kw["on"] = "RESTRICT|CASCADE|SET NULL|NO ACTION"
        self._re_fk_constraint = _re_compile(
            r"  "
            r"CONSTRAINT +"
            r"{iq}(?P<name>(?:{esc_fq}|[^{fq}])+){fq} +"
            r"FOREIGN KEY +"
            r"\((?P<local>[^\)]+?)\) REFERENCES +"
            r"(?P<table>{iq}[^{fq}]+{fq}"
            r"(?:\.{iq}[^{fq}]+{fq})?) *"
            r"\((?P<foreign>(?:{iq}[^{fq}]+{fq}(?: *, *)?)+)\)"
            r"(?: +(?P<match>MATCH \w+))?"
            r"(?: +ON UPDATE (?P<onupdate>{on}))?"
            r"(?: +ON DELETE (?P<ondelete>{on}))?".format(
                iq=quotes["iq"], esc_fq=quotes["esc_fq"], fq=quotes["fq"], on=kw["on"]
            )
        )

    def _parse_constraints(self, line):
        """Parse a CONSTRAINT line."""
        ret = super()._parse_constraints(line)
        if ret:
            tp, spec = ret
            if tp == "partition":
                # do not handle partition
                return ret
            # logger.info(f"{tp} {spec}")
            if (
                tp == "fk_constraint"
                and len(spec["table"]) == 2
                and spec["table"][0] == self.default_schema
            ):
                spec["table"] = spec["table"][1:]
            if spec.get("onupdate", "").lower() == "restrict":
                spec["onupdate"] = None
            if spec.get("ondelete", "").lower() == "restrict":
                spec["ondelete"] = None
        return ret


class OBDialect(pymysql.MySQLDialect_pymysql):
    """OBDialect expend."""

    supports_statement_cache = True

    def __init__(self, **kwargs):
        """Initialize OBDialect."""
        try:
            from pyobvector import VECTOR  # type: ignore
        except ImportError:
            raise ImportError(
                "Could not import pyobvector package. "
                "Please install it with `pip install pyobvector`."
            )
        super().__init__(**kwargs)
        self.ischema_names["VECTOR"] = VECTOR

    def initialize(self, connection):
        """Ob dialect initialize."""
        super(OBDialect, self).initialize(connection)
        self._server_version_info = (5, 7, 19)
        self.server_version_info = (5, 7, 19)

    def _server_version_info(self, connection):
        """Ob set fixed version ending compatibility issue."""
        return (5, 7, 19)

    def get_isolation_level(self, dbapi_connection):
        """Ob set fixed version ending compatibility issue."""
        self.server_version_info = (5, 7, 19)
        return super(OBDialect, self).get_isolation_level(dbapi_connection)

    @util.memoized_property
    def _tabledef_parser(self):
        """Return the MySQLTableDefinitionParser, generate if needed.

        The deferred creation ensures that the dialect has
        retrieved server version information first.
        """
        preparer = self.identifier_preparer
        default_schema = self.default_schema_name
        return OceanBaseTableDefinitionParser(
            self, preparer, default_schema=default_schema
        )


registry.register("mysql.ob", __name__, "OBDialect")



"""TuGraph store."""

import base64
import json
import logging
import os
from dataclasses import dataclass, field
from typing import List

from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.graph_store.base import GraphStoreBase, GraphStoreConfig
from dbgpt.storage.graph_store.graph import GraphElemType
from dbgpt.util.i18n_utils import _
from dbgpt_ext.datasource.conn_tugraph import TuGraphConnector

logger = logging.getLogger(__name__)


@register_resource(
    _("TuGraph Graph Config"),
    "tugraph_config",
    category=ResourceCategory.KNOWLEDGE_GRAPH,
    description=_("TuGraph config."),
    parameters=[
        Parameter.build_from(
            _("host"),
            "host",
            str,
            optional=True,
            default="127.0.0.1",
            description=_("TuGraph host"),
        ),
        Parameter.build_from(
            _("port"),
            "port",
            int,
            optional=True,
            default="7687",
            description=_("TuGraph port"),
        ),
        Parameter.build_from(
            _("username"),
            "username",
            str,
            optional=True,
            default="admin",
            description=_("TuGraph username"),
        ),
        Parameter.build_from(
            _("password"),
            "password",
            str,
            optional=True,
            default="73@TuGraph",
            description=_("TuGraph password"),
        ),
    ],
)
@dataclass
class TuGraphStoreConfig(GraphStoreConfig):
    """TuGraph store config."""

    __type__ = "tugraph"

    host: str = field(
        default="127.0.0.1",
        metadata={
            "description": "TuGraph host",
        },
    )
    port: int = field(
        default=7687,
        metadata={
            "description": "TuGraph port",
        },
    )
    username: str = field(
        default="admin",
        metadata={
            "description": "login username",
        },
    )
    password: str = field(
        default="73@TuGraph",
        metadata={
            "description": "login password",
        },
    )
    vertex_type: str = field(
        default=GraphElemType.ENTITY.value,
        metadata={
            "description": "The type of vertex, `entity` by default.",
        },
    )
    document_type: str = field(
        default=GraphElemType.DOCUMENT.value,
        metadata={
            "description": "The type of document vertex, `document` by default.",
        },
    )
    chunk_type: str = field(
        default=GraphElemType.CHUNK.value,
        metadata={
            "description": "The type of chunk vertex, `relation` by default.",
        },
    )
    edge_type: str = field(
        default=GraphElemType.RELATION.value,
        metadata={
            "description": "The type of relation edge, `relation` by default.",
        },
    )
    include_type: str = field(
        default=GraphElemType.INCLUDE.value,
        metadata={
            "description": "The type of include edge, `include` by default.",
        },
    )
    next_type: str = field(
        default=GraphElemType.NEXT.value,
        metadata={
            "description": "The type of next edge, `next` by default.",
        },
    )
    plugin_names: List[str] = field(
        default_factory=lambda: ["leiden"],
        metadata={
            "description": "The list of plugin names to be uploaded to the database.",
        },
    )
    enable_summary: bool = field(
        default=True,
        metadata={
            "description": "Enable graph community summary or not.",
        },
    )
    enable_similarity_search: bool = field(
        default=False,
        metadata={
            "description": "Enable the similarity search or not",
        },
    )


class TuGraphStore(GraphStoreBase):
    """TuGraph graph store."""

    def __init__(self, config: TuGraphStoreConfig) -> None:
        """Initialize the TuGraphStore with connection details."""
        self._config = config
        self._host = config.host or os.getenv("TUGRAPH_HOST")
        self._port = int(config.port or os.getenv("TUGRAPH_PORT"))
        self._username = config.username or os.getenv("TUGRAPH_USERNAME")
        self._password = config.password or os.getenv("TUGRAPH_PASSWORD")
        self.enable_summary = config.enable_summary or (
            os.getenv("GRAPH_COMMUNITY_SUMMARY_ENABLED", "").lower() == "true"
        )
        self.enable_similarity_search = config.enable_similarity_search or (
            os.getenv("SIMILARITY_SEARCH_ENABLED", "").lower() == "true"
        )
        self._plugin_names = config.plugin_names or (
            os.getenv("TUGRAPH_PLUGIN_NAMES", "leiden").split(",")
        )

        self._graph_name = config.name

        self.conn = TuGraphConnector.from_uri_db(
            host=self._host,
            port=self._port,
            user=self._username,
            pwd=self._password,
            db_name=config.name,
        )

    def get_config(self) -> TuGraphStoreConfig:
        """Get the TuGraph store config."""
        return self._config

    def is_exist(self, name) -> bool:
        """Check Graph Name is Exist."""
        return self.conn.is_exist(name)

    def _add_vertex_index(self, field_name):
        """Add an index to the vertex table."""
        # TODO: Not used in the current implementation.
        gql = f"CALL db.addIndex('{GraphElemType.ENTITY.value}', '{field_name}', false)"
        self.conn.run(gql)

    def _upload_plugin(self):
        """Upload missing plugins to the TuGraph database.

        This method checks for the presence of required plugins in the database and
        uploads any missing plugins. It performs the following steps:
        1. Lists existing plugins in the database.
        2. Identifies missing plugins by comparing with the required plugin list.
        3. For each missing plugin, reads its binary content, encodes it, and uploads to
            the database.

        The method uses the 'leiden' plugin as an example, but can be extended for other
        plugins.
        """
        gql = "CALL db.plugin.listPlugin('CPP','v1')"
        result = self.conn.run(gql)
        result_names = [
            json.loads(record["plugin_description"])["name"] for record in result
        ]
        missing_plugins = [
            name for name in self._plugin_names if name not in result_names
        ]

        if len(missing_plugins):
            for name in missing_plugins:
                try:
                    from dbgpt_tugraph_plugins import (  # type: ignore
                        get_plugin_binary_path,
                    )
                except ImportError:
                    logger.error(
                        "dbgpt-tugraph-plugins is not installed, "
                        "pip install dbgpt-tugraph-plugins==0.1.1"
                    )
                plugin_path = get_plugin_binary_path("leiden")  # type: ignore
                with open(plugin_path, "rb") as f:
                    content = f.read()
                content = base64.b64encode(content).decode()
                gql = (
                    f"CALL db.plugin.loadPlugin('CPP', '{name}', '{content}', 'SO', "
                    f"'{name} Plugin', false, 'v1')"
                )
                self.conn.run(gql)



import struct
from typing import Any, Dict, List, Optional, Union

import olefile

from dbgpt.core import Document
from dbgpt.rag.knowledge.base import (
    ChunkStrategy,
    DocumentType,
    Knowledge,
    KnowledgeType,
)


class Word97DocParser:
    """Parser for Microsoft Word 97-2003 (.doc) binary files.

    This module implements a parser for the legacy Word 97-2003 binary format (.doc),
    based on the official Microsoft [MS-DOC] specification.

    Specification Reference:
        [MS-DOC]: Word (.doc) Binary File Format
        https://learn.microsoft.com/en-us/openspecs/office_file_formats/ms-doc/ccd7b486-7881-484c-a137-51170af7cc22

    Example:
        >>> from doc import Word97DocParser
        >>> with Word97DocParser("example.doc") as parser:
        ...     paragraphs = parser.extract_text_by_paragraphs()
        ...     for i, para in enumerate(paragraphs, 1):
        ...         print(f"\nParagraph {i}:")
        ...         print(para)
    """

    # Mapping of special ANSI characters to Unicode
    ANSI_TO_UNICODE = {
        0x82: 0x201A,
        0x83: 0x0192,
        0x84: 0x201E,
        0x85: 0x2026,
        0x86: 0x2020,
        0x87: 0x2021,
        0x88: 0x02C6,
        0x89: 0x2030,
        0x8A: 0x0160,
        0x8B: 0x2039,
        0x8C: 0x0152,
        0x91: 0x2018,
        0x92: 0x2019,
        0x93: 0x201C,
        0x94: 0x201D,
        0x95: 0x2022,
        0x96: 0x2013,
        0x97: 0x2014,
        0x98: 0x02DC,
        0x99: 0x2122,
        0x9A: 0x0161,
        0x9B: 0x203A,
        0x9C: 0x0153,
        0x9F: 0x0178,
    }

    # Mapping of nFib values to cbRgFcLcb sizes
    FIB_VERSIONS = {
        0x00C1: 0x005D,  # Word 97
        0x00D9: 0x006C,  # Word 2000
        0x0101: 0x0088,  # Word 2002
        0x010C: 0x00A4,  # Word 2003
        0x0112: 0x00B7,  # Word 2007
    }

    def __init__(self, doc_path):
        """Initialize with path to Word document"""
        self.doc_path = doc_path
        self.ole = None
        self.word_doc_stream = None
        self.table_stream = None
        self.fib_info = None
        self.plc_pcd = None

    def __enter__(self):
        """Context manager entry"""
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit"""
        self.close()

    def open(self):
        """Open the OLE file and required streams"""
        self.ole = olefile.OleFileIO(self.doc_path)

        if not self.ole.exists("WordDocument"):
            raise ValueError("WordDocument stream not found")

        self.word_doc_stream = self.ole.openstream("WordDocument")

        # Determine table stream name
        table_stream_name = "0Table" if self.ole.exists("0Table") else "1Table"
        self.table_stream = self.ole.openstream(table_stream_name)

    def close(self):
        """Close all open resources"""
        if self.ole:
            self.ole.close()
        self.ole = None
        self.word_doc_stream = None
        self.table_stream = None

    def read_fib(self):
        """Read the File Information Block (FIB) from the WordDocument stream (2.5)"""
        # Read FibBase (32 bytes)
        fib_base = self.word_doc_stream.read(32)

        # Unpack nFib (offset 2, size 2 bytes)
        nFib = struct.unpack("<H", fib_base[2:4])[0]
        if nFib not in self.FIB_VERSIONS:
            raise ValueError(f"Unsupported nFib version: 0x{nFib:04X}")

        # Skip csw (2 bytes at offset 32)
        self.word_doc_stream.read(2)

        # Skip FibRgW97 (28 bytes) and cslw (2 bytes)
        self.word_doc_stream.read(28 + 2)

        # Read FibRgLw97 (88 bytes)
        fib_rg_lw = self.word_doc_stream.read(88)
        ccpText = struct.unpack("<I", fib_rg_lw[12:16])[0]  # Total character count

        # Read cbRgFcLcb (2 bytes)
        cb_rg_fc_lcb = struct.unpack("<H", self.word_doc_stream.read(2))[0]

        # Read FibRgFcLcbBlob (variable size)
        fib_rg_fc_lcb_blob = self.word_doc_stream.read(cb_rg_fc_lcb * 8)

        # Skip cswNew (2 bytes) and FibRgCswNew (variable size)
        csw_new = struct.unpack("<H", self.word_doc_stream.read(2))[0]
        if csw_new > 0:
            self.word_doc_stream.read(csw_new * 2)

        # Extract fcClx and lcbClx from FibRgFcLcb97
        fc_clx = struct.unpack("<I", fib_rg_fc_lcb_blob[0x108:0x10C])[0]
        lcb_clx = struct.unpack("<I", fib_rg_fc_lcb_blob[0x10C:0x110])[0]
        fc_plcf_bte_papx = struct.unpack("<I", fib_rg_fc_lcb_blob[0x68:0x6C])[0]
        lcb_plcf_bte_papx = struct.unpack("<I", fib_rg_fc_lcb_blob[0x6C:0x70])[0]
        self.fib_info = {
            "nFib": nFib,
            "fcClx": fc_clx,
            "lcbClx": lcb_clx,
            "ccpText": ccpText,
            "fcPlcfBtePapx": fc_plcf_bte_papx,
            "lcbPlcfBtePapx": lcb_plcf_bte_papx,
        }
        return self.fib_info

    def read_clx(self, fc_clx, lcb_clx):
        """Read the CLX structure from the Table stream"""
        self.table_stream.seek(fc_clx)
        clx_data = self.table_stream.read(lcb_clx)

        # For simplicity, we assume the data starts with Pcdt (0x02)
        if clx_data[0] != 0x02:
            raise ValueError("Expected Pcdt structure not found in CLX data")

        return clx_data

    def parse_plc_pcd(self, pcdt_data):
        """Parse the PLC structure containing PCDs(2.9.177)"""
        # Get the size of PlcPcd structure
        lcb = struct.unpack("<I", pcdt_data[1:5])[0]
        plc_pcd_bytes = pcdt_data[5 : 5 + lcb]

        # Calculate number of PCDs: n = (lcb - 4) // 12
        n = (lcb - 4) // 12

        # Parse aCP array (n+1 CPs, each 4 bytes)
        aCP = [
            struct.unpack("<I", plc_pcd_bytes[i * 4 : (i + 1) * 4])[0]
            for i in range(n + 1)
        ]

        # Parse aPcd array (n PCDs, each 8 bytes)
        aPcd = []
        for i in range(n):
            start = (n + 1) * 4 + i * 8
            pcd_bytes = plc_pcd_bytes[start : start + 8]

            # Extract fc (bytes 2-6) and compression flag
            fc_bytes = pcd_bytes[2:6]
            fc = int.from_bytes(fc_bytes, byteorder="little", signed=False)
            fcompressed = (fc >> 1) & 0x1

            aPcd.append(
                {
                    "fc": fc,
                    "start": aCP[i],
                    "end": aCP[i + 1],
                    "f_compressed": fcompressed,
                }
            )
        self.plc_pcd = {"aCP": aCP, "aPcd": aPcd}
        return self.plc_pcd

    def _find_pcd_index(self, cp):
        """Find the index of the PCD containing the given character position"""
        aCP = self.plc_pcd["aCP"]
        for i in range(len(aCP) - 1):
            if aCP[i] <= cp < aCP[i + 1]:
                return i
        return None

    def get_paragraph_boundaries(self, cp):
        """Find paragraph boundaries for a given character position (2.4.2)"""
        if not self.fib_info or not self.plc_pcd:
            raise RuntimeError("Must read FIB and PLC/PCD first")

        # Find the PCD containing this cp
        i = self._find_pcd_index(cp)
        if i is None:
            return None

        pcd = self.plc_pcd["aPcd"][i]
        start_cp = self._find_paragraph_start(cp, i, pcd)
        end_cp = self._find_paragraph_end(cp, i, pcd)

        return (start_cp, end_cp)

    def _find_paragraph_start(self, cp, i, pcd):
        """Find the start of the paragraph containing cp (algorithm 2.4.2)"""
        # Step 3: Calculate fc and fc_pcd
        fc_pcd = pcd["fc"]
        # Let fcPcd be Pcd.fc.fc. Let fc be fcPcd + 2(cp – PlcPcd.aCp[i]).
        # If Pcd.fc.fCompressed is one, set fc to fc / 2, and set fcPcd to fcPcd/2.
        fc = fc_pcd + 2 * (cp - self.plc_pcd["aCP"][i])
        if pcd["f_compressed"]:
            fc = fc // 2
            fc_pcd = fc_pcd // 2

        # Step 4: Read PlcBtePapx
        self.table_stream.seek(self.fib_info["fcPlcfBtePapx"])
        plcf_bte_papx_data = self.table_stream.read(self.fib_info["lcbPlcfBtePapx"])
        a_fc, a_pn = self._parse_plcf_bte_papx(plcf_bte_papx_data)

        # Handle case where a_fc is empty
        if not a_fc:
            return None

        fc_last = a_fc[-1]

        # Step 4 continued: Check fcLast
        if fc_last <= fc:
            if fc_last < fc_pcd:
                # Step 8: Check if at beginning of document
                if self.plc_pcd["aCP"][i] == 0:
                    return 0
                # Step 9: Recurse with previous cp
                return self._find_paragraph_start(
                    self.plc_pcd["aCP"][i], i - 1, self.plc_pcd["aPcd"][i - 1]
                )
            # Adjust fc and fc_last if needed
            fc = fc_last
            if pcd["f_compressed"]:
                fc_last = fc_last // 2
            fc_first = fc_last
        else:
            # Step 5: Find largest j where a_fc[j] <= fc
            j = self._find_largest_index_le(a_fc, fc)
            if j is None:
                return None  # Invalid cp

            # Read PapxFkp
            papx_fkp = self._read_papx_fkp(a_pn[j])
            # print(f"papx_fkp:{papx_fkp}, j:{j}")
            if not papx_fkp or not papx_fkp.get("rgfc"):
                return None  # Invalid data

            # Step 6: Find largest k where rgfc[k] <= fc
            k = self._find_largest_index_le(papx_fkp["rgfc"], fc)
            if k is None:
                return None  # Invalid cp

            # Check if cp is outside document range
            if papx_fkp["rgfc"][-1] <= fc:
                return None

            fc_first = papx_fkp["rgfc"][k]

        # Step 7: Calculate paragraph start
        if fc_first > fc_pcd:
            dfc = fc_first - fc_pcd
            if not pcd["f_compressed"]:
                dfc = dfc // 2
            return self.plc_pcd["aCP"][i] + dfc

        # Step 8: Check if at beginning of document
        if self.plc_pcd["aCP"][i] == 0:
            return 0

        # Step 9: Recurse with previous cp
        return self._find_paragraph_start(
            self.plc_pcd["aCP"][i], i - 1, self.plc_pcd["aPcd"][i - 1]
        )

    def _find_paragraph_end(self, cp, i, pcd):
        """Find the end of the paragraph containing cp (algorithm 2.4.2)"""
        fc_pcd = pcd["fc"]
        fc = fc_pcd + 2 * (cp - self.plc_pcd["aCP"][i])
        fc_mac = fc_pcd + 2 * (self.plc_pcd["aCP"][i + 1] - self.plc_pcd["aCP"][i])

        if pcd["f_compressed"]:
            fc = fc // 2
            fc_pcd = fc_pcd // 2
            fc_mac = fc_mac // 2

        # Read PlcBtePapx
        self.table_stream.seek(self.fib_info["fcPlcfBtePapx"])
        plcf_bte_papx_data = self.table_stream.read(self.fib_info["lcbPlcfBtePapx"])
        a_fc, a_pn = self._parse_plcf_bte_papx(plcf_bte_papx_data)

        # Find largest j where a_fc[j] <= fc
        j = self._find_largest_index_le(a_fc, fc)
        if j is None or (a_fc and fc >= a_fc[-1]):
            return self._find_paragraph_end(
                self.plc_pcd["aCP"][i + 1], i + 1, self.plc_pcd["aPcd"][i + 1]
            )

        # Read PapxFkp
        papx_fkp = self._read_papx_fkp(a_pn[j])
        if not papx_fkp:
            return None

        # Find largest k where rgfc[k] <= fc
        k = self._find_largest_index_le(papx_fkp["rgfc"], fc)
        if k is None or (papx_fkp["rgfc"] and fc >= papx_fkp["rgfc"][-1]):
            return None

        fc_lim = papx_fkp["rgfc"][k + 1] if k + 1 < len(papx_fkp["rgfc"]) else fc_mac

        if fc_lim <= fc_mac:
            dfc = fc_lim - fc_pcd
            if not pcd["f_compressed"]:
                dfc = dfc // 2
            return self.plc_pcd["aCP"][i] + dfc - 1

        return self._find_paragraph_end(
            self.plc_pcd["aCP"][i + 1], i + 1, self.plc_pcd["aPcd"][i + 1]
        )

    def _parse_plcf_bte_papx(self, data):
        """Parse PlcBtePapx structure (2.8.6)

        Args:
            data: Raw bytes of PlcBtePapx structure.

        Returns:
            (a_fc, a_pn): Tuple of two lists:
                - a_fc: List of unsigned 4-byte integers (FC offsets).
                - a_pn: List of unsigned 4-byte integers (PnFkpPapx entries).

        Raises:
            ValueError: If data is malformed or aFC is not sorted/unique.
        """
        if len(data) < 12:  # Minimum: 8 (aFC[0..1]) + 4 (aPnBtePapx[0])
            return [], []

        # Calculate number of aPnBtePapx entries (n)
        n = (len(data) - 4) // 8
        if (2 * n + 1) * 4 != len(data):
            raise ValueError("Invalid PlcBtePapx size")

        a_fc = []
        a_pn = []

        # Parse aFC (n+1 entries, each 4 bytes)
        for i in range(n + 1):
            offset = i * 4
            fc = struct.unpack("<I", data[offset : offset + 4])[0]
            a_fc.append(fc)

        # Parse aPnBtePapx (n entries, each 4 bytes, starting after last aFC)
        pn_offset = (n + 1) * 4
        for i in range(n):
            offset = pn_offset + i * 4
            pn = struct.unpack("<I", data[offset : offset + 4])[0]
            a_pn.append(pn)

        # Validate aFC is strictly increasing (sorted and unique)
        for i in range(len(a_fc) - 1):
            if a_fc[i] >= a_fc[i + 1]:
                raise ValueError("aFC must be strictly increasing")

        return a_fc, a_pn

    def _read_papx_fkp(self, pn):
        """Read PapxFkp structure from WordDocument stream.

        Args:
            pn: Page number (PnFkpPapx), offset = pn * 512.

        Returns:
            Dict with keys:
                - "rgfc": List of FC offsets (4-byte unsigned integers).
                - "rgbx": List of BxPap (1-byte integers).
                - "papx_in_fkp": List of PapxInFkp raw bytes.

        Raises:
            ValueError: If FKP data is invalid.
        """
        offset = pn * 512
        self.word_doc_stream.seek(offset)
        fkp_data = self.word_doc_stream.read(512)

        if len(fkp_data) != 512:
            raise ValueError("FKP size must be 512 bytes")

        cpara = fkp_data[511]  # Number of paragraphs (1 ≤ cpara ≤ 0x1D)
        if not 1 <= cpara <= 0x1D:
            raise ValueError(f"Invalid cpara: {cpara} (must be 1 ≤ cpara ≤ 29)")

        # Parse rgfc (cpara + 1 entries, each 4 bytes)
        rgfc = []
        for i in range(cpara + 1):
            fc_offset = i * 4
            fc = struct.unpack("<I", fkp_data[fc_offset : fc_offset + 4])[0]
            rgfc.append(fc)

        # Parse rgbx (cpara entries, each 1 byte)
        rgbx_start = (cpara + 1) * 4
        rgbx = list(fkp_data[rgbx_start : rgbx_start + cpara])

        # Parse PapxInFkp (variable size, located after rgbx)
        papx_in_fkp_start = rgbx_start + cpara
        papx_in_fkp_end = 511  # cpara is the last byte
        papx_in_fkp = fkp_data[papx_in_fkp_start:papx_in_fkp_end]

        return {
            "rgfc": rgfc,
            "rgbx": rgbx,
            "papx_in_fkp": papx_in_fkp,
        }

    def _find_largest_index_le(self, array, value):
        """Find largest index where array[index] <= value"""
        for i in reversed(range(len(array))):
            if array[i] <= value:
                return i
        return None

    def extract_text_by_paragraphs(self):
        """Extract text organized by paragraphs"""
        self.word_doc_stream.seek(0)
        self.table_stream.seek(0)
        self.read_fib()
        clx_data = self.read_clx(self.fib_info["fcClx"], self.fib_info["lcbClx"])
        self.parse_plc_pcd(clx_data)

        paragraphs = []
        current_cp = 0
        total_chars = self.fib_info["ccpText"]

        while current_cp < total_chars:
            boundaries = self.get_paragraph_boundaries(current_cp)
            if not boundaries:
                break

            start, end = boundaries
            if start > end:
                break

            paragraph_text = self._extract_text_range(start, end)
            paragraphs.append(paragraph_text)
            current_cp = end + 1

        return paragraphs

    def _extract_text_range(self, start_cp, end_cp):
        """Extract text between two character positions"""
        text_chars = []
        i = self._find_pcd_index(start_cp)

        while i is not None and start_cp <= end_cp:
            pcd = self.plc_pcd["aPcd"][i]
            pcd_start = self.plc_pcd["aCP"][i]
            pcd_end = self.plc_pcd["aCP"][i + 1]

            # Determine range within this PCD
            range_start = max(start_cp, pcd_start)
            range_end = min(end_cp, pcd_end - 1)

            if range_start > range_end:
                i += 1
                continue

            fc = pcd["fc"]
            compressed = pcd["f_compressed"]

            for cp in range(range_start, range_end + 1):
                if compressed:
                    offset = fc + (cp - pcd_start)
                    self.word_doc_stream.seek(offset)
                    char_byte = self.word_doc_stream.read(1)
                    char_code = char_byte[0]
                    char = chr(self.ANSI_TO_UNICODE.get(char_code, char_code))
                else:
                    offset = fc + 2 * (cp - pcd_start)
                    self.word_doc_stream.seek(offset)
                    char_bytes = self.word_doc_stream.read(2)
                    char = char_bytes.decode("utf-16-le")

                text_chars.append(char)

            start_cp = range_end + 1
            i += 1

        return "".join(text_chars)

    def extract_text(self):
        """Main method to extract text from the document"""
        self.word_doc_stream.seek(0)
        self.table_stream.seek(0)
        fib_info = self.read_fib()
        clx_data = self.read_clx(fib_info["fcClx"], fib_info["lcbClx"])
        pcd_array = self.parse_plc_pcd(clx_data)

        full_text = []
        for pcd in pcd_array["aPcd"]:
            start_cp, end_cp = pcd["start"], pcd["end"]
            char_count = end_cp - start_cp

            if char_count == 0:
                continue

            fc = pcd["fc"]
            compressed = pcd["f_compressed"]
            text_chars = []

            for cp in range(start_cp, end_cp):
                offset = (
                    fc + (cp - start_cp) if compressed else fc + 2 * (cp - start_cp)
                )
                self.word_doc_stream.seek(offset)

                if compressed:
                    char_byte = self.word_doc_stream.read(1)
                    char_code = char_byte[0]
                    char = chr(self.ANSI_TO_UNICODE.get(char_code, char_code))
                else:
                    char_bytes = self.word_doc_stream.read(2)
                    # decode char
                    char = char_bytes.decode("utf-16-le")
                text_chars.append(char)
            full_text.append("".join(text_chars))

        return "".join(full_text)


class Word97DocKnowledge(Knowledge):
    """Microsoft Word 97-2003 (.doc)."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        knowledge_type: Any = KnowledgeType.DOCUMENT,
        encoding: Optional[str] = "utf-16-le",
        loader: Optional[Any] = None,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
        **kwargs: Any,
    ) -> None:
        """Create  Microsoft Word 97-2003 (.doc) Knowledge with Knowledge arguments.

        Args:
            file_path(str,  optional): file path
            knowledge_type(KnowledgeType, optional): knowledge type
            encoding(str, optional): .doc encoding
            loader(Any, optional): loader
        """
        super().__init__(
            path=file_path,
            knowledge_type=knowledge_type,
            data_loader=loader,
            metadata=metadata,
            **kwargs,
        )
        self._encoding = encoding

    def _load(self) -> List[Document]:
        """Load doc document from loader."""
        if self._loader:
            documents = self._loader.load()
        else:
            docs = []
            content = []
            with Word97DocParser(self._path) as parser:
                paragraphs = parser.extract_text_by_paragraphs()
                for i, para in enumerate(paragraphs):
                    content.append(para)

            metadata = {"source": self._path}
            if self._metadata:
                metadata.update(self._metadata)  # type: ignore
            docs.append(Document(content="\n".join(content), metadata=metadata))
            return docs
        return [Document.langchain2doc(lc_document) for lc_document in documents]

    @classmethod
    def support_chunk_strategy(cls) -> List[ChunkStrategy]:
        """Return support chunk strategy."""
        return [
            ChunkStrategy.CHUNK_BY_SIZE,
            ChunkStrategy.CHUNK_BY_PARAGRAPH,
            ChunkStrategy.CHUNK_BY_SEPARATOR,
        ]

    @classmethod
    def default_chunk_strategy(cls) -> ChunkStrategy:
        """Return default chunk strategy."""
        return ChunkStrategy.CHUNK_BY_SIZE

    @classmethod
    def type(cls) -> KnowledgeType:
        """Return knowledge type."""
        return KnowledgeType.DOCUMENT

    @classmethod
    def document_type(cls) -> DocumentType:
        """Return document type."""
        return DocumentType.DOC



"""Base Assembler."""

from abc import ABC, abstractmethod
from typing import Any, List, Optional

from dbgpt.core import Chunk
from dbgpt.rag.knowledge.base import Knowledge
from dbgpt.rag.retriever.base import BaseRetriever
from dbgpt.rag.transformer.base import ExtractorBase
from dbgpt.util.tracer import root_tracer

from ..chunk_manager import ChunkManager, ChunkParameters


class BaseAssembler(ABC):
    """Base Assembler."""

    def __init__(
        self,
        knowledge: Knowledge,
        chunk_parameters: Optional[ChunkParameters] = None,
        extractor: Optional[ExtractorBase] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize with Assembler arguments.

        Args:
            knowledge(Knowledge): Knowledge datasource.
            chunk_parameters: (Optional[ChunkParameters]) ChunkManager to use for
                chunking.
            extractor(Optional[ExtractorBase]):  ExtractorBase to use for summarization.
        """
        self._knowledge = knowledge
        self._chunk_parameters = chunk_parameters or ChunkParameters()
        self._extractor = extractor
        self._chunk_manager = ChunkManager(
            knowledge=self._knowledge, chunk_parameter=self._chunk_parameters
        )
        self._chunks: List[Chunk] = []
        metadata = {
            "knowledge_cls": (
                self._knowledge.__class__.__name__ if self._knowledge else None
            ),
            "knowledge_type": self._knowledge.type().value if self._knowledge else None,
            "path": (
                self._knowledge._path
                if self._knowledge and hasattr(self._knowledge, "_path")
                else None
            ),
            "chunk_parameters": self._chunk_parameters.dict(),
        }
        with root_tracer.start_span("BaseAssembler.load_knowledge", metadata=metadata):
            self.load_knowledge(self._knowledge)

    def load_knowledge(self, knowledge: Knowledge) -> None:
        """Load knowledge Pipeline."""
        if not knowledge:
            raise ValueError("knowledge must be provided.")
        with root_tracer.start_span("BaseAssembler.knowledge.load"):
            documents = knowledge.load()
        with root_tracer.start_span("BaseAssembler.chunk_manager.split"):
            self._chunks = self._chunk_manager.split(documents)

    @abstractmethod
    def as_retriever(self, **kwargs: Any) -> BaseRetriever:
        """Return a retriever."""

    @abstractmethod
    def persist(self, **kwargs: Any) -> List[str]:
        """Persist chunks.

        Returns:
            List[str]: List of persisted chunk ids.
        """

    async def apersist(self, **kwargs: Any) -> List[str]:
        """Persist chunks.

        Returns:
            List[str]: List of persisted chunk ids.
        """
        raise NotImplementedError

    def get_chunks(self) -> List[Chunk]:
        """Return chunks."""
        return self._chunks



"""Keyword Based Graph Retriever."""

import logging
from typing import List, Tuple

from dbgpt.storage.graph_store.graph import Graph
from dbgpt_ext.rag.retriever.graph_retriever.base import GraphRetrieverBase

logger = logging.getLogger(__name__)


class KeywordBasedGraphRetriever(GraphRetrieverBase):
    """Keyword Based Graph Retriever class."""

    def __init__(self, graph_store_adapter, triplet_topk):
        """Initialize Keyword Based Graph Retriever."""
        self._graph_store_adapter = graph_store_adapter
        self._triplet_topk = triplet_topk

    async def retrieve(self, keywords: List[str]) -> Tuple[Graph, str]:
        """Retrieve from triplets graph with keywords."""
        subgraph = self._graph_store_adapter.explore_trigraph(
            subs=keywords,
            limit=self._triplet_topk,
        )

        return subgraph
